import deal


@deal.reason(ZeroDivisionError, lambda a, b: b == 0)
@deal.reason(ValueError, lambda a, b: a == b, message='a is equal to b')
@deal.raises(ValueError, IndexError, ZeroDivisionError)
@deal.pre(lambda a, b: b != 0)
@deal.pre(lambda a, b: b != 0, message='b is not zero')
@deal.ensure(lambda a, b, result: b != result)
@deal.post(lambda res: res != .13)
@deal.has('database', 'network')
@deal.example(lambda: example(6, 2) == 3)
def example(a: int, b: int) -> float:
    """Example function.

    :return: The description for return value.
    """
    return a / b



import typing
from functools import update_wrapper
from inspect import signature

import hypothesis
import hypothesis.strategies
import typeguard
from hypothesis.internal.reflection import proxies

from . import introspection
from ._cached_property import cached_property


F = typing.Callable[..., None]
FuzzInputType = typing.Union[bytes, bytearray, memoryview, typing.BinaryIO]
FuzzType = typing.Callable[[FuzzInputType], typing.Optional[bytes]]
ArgsKwargsType = typing.Tuple[tuple, typing.Dict[str, typing.Any]]
EXAMPLE = object()


class TestCase(typing.NamedTuple):
    """A callable object, wrapper around a function that must be tested.

    When called, calls the wrapped function, suppresses expected exceptions,
    checks the type of the result, and returns it.
    """

    args: typing.Tuple[typing.Any, ...]
    """Positional arguments to be passed in the function"""

    kwargs: typing.Dict[str, typing.Any]
    """Keyword arguments to be passed in the function"""

    func: typing.Callable
    """The function which will be called when the test case is called"""

    exceptions: typing.Tuple[typing.Type[Exception], ...]
    """Exceptions that must be suppressed.
    """

    check_types: bool
    """Check that the result matches return type of the function.
    """

    def __call__(self) -> typing.Any:
        """Calls the given test case returning the called functions result on success or
        Raising an exception on error
        """
        __tracebackhide__ = True
        try:
            result = self.func(*self.args, **self.kwargs)
        except self.exceptions:
            return typing.NoReturn
        self._check_result(result)
        return result

    def _check_result(self, result: typing.Any) -> None:
        if not self.check_types:
            return
        memo = typeguard._CallMemo(
            func=self.func,
            args=self.args,
            kwargs=self.kwargs,
        )
        typeguard.check_argument_types(memo=memo)
        typeguard.check_return_type(result, memo=memo)


class cases:  # noqa: N
    """Generate test cases for the given function.
    """

    func: typing.Callable
    """the function to test. Should be type annotated."""

    count: int
    """how many test cases to generate, defaults to 50."""

    kwargs: typing.Dict[str, typing.Any]
    """keyword arguments to pass into the function."""

    check_types: bool
    """check that the result matches return type of the function. Enabled by default."""

    settings: hypothesis.settings
    """Hypothesis settings to use instead of default ones."""

    seed: typing.Optional[int]
    """Random seed to use when generating test cases. Use it to make tests deterministic."""

    def __init__(
        self,
        func: typing.Callable, *,
        count: int = 50,
        kwargs: typing.Optional[typing.Dict[str, typing.Any]] = None,
        check_types: bool = True,
        settings: typing.Optional[hypothesis.settings] = None,
        seed: typing.Optional[int] = None,
    ) -> None:
        """
        Create test cases generator.

        ```pycon
        >>> import deal
        >>> @deal.pre(lambda a, b: b != 0)
        ... def div(a: int, b: int) -> float:
        ...   return a / b
        ...
        >>> cases = deal.cases(div)
        >>>
        ```

        """
        self.func = func  # type: ignore
        self.count = count
        self.kwargs = kwargs or {}
        self.check_types = check_types
        self.settings = settings or self._default_settings
        self.seed = seed

    def __iter__(self) -> typing.Iterator[TestCase]:
        """Emits test cases.

        It can be helpful when you want to see what test cases are generated.
        The recommend way is to use `deal.cases` as a decorator instead.

        ```pycon
        >>> import deal
        >>> @deal.pre(lambda a, b: b != 0)
        ... def div(a: int, b: int) -> float:
        ...   return a / b
        ...
        >>> cases = iter(deal.cases(div))
        >>> next(cases)
        TestCase(args=(), kwargs=..., func=<function div ...>, exceptions=(), check_types=True)
        >>> for case in cases:
        ...   result = case()  # execute the test case
        >>>
        ```

        """
        cases: typing.List[TestCase] = []
        test = self(cases.append)
        test()
        yield from cases

    def __repr__(self) -> str:
        args = [
            getattr(self.func, '__name__', repr(self.func)),
            f'count={self.count}',
        ]
        if self.seed is not None:
            args.append(f'seed={self.seed}')
        if self.kwargs:
            args.append(f'kwargs={repr(self.kwargs)}')
        return 'deal.cases({})'.format(', '.join(args))

    def _make_case(self, *args, **kwargs) -> TestCase:
        """Make test case with the given arguments.
        """
        return TestCase(
            args=args,
            kwargs=kwargs,
            func=self.func,
            exceptions=self.exceptions,
            check_types=self.check_types,
        )

    @cached_property
    def _contracts(self) -> typing.Tuple[introspection.Contract, ...]:
        return tuple(introspection.get_contracts(self.func))

    @cached_property
    def _pres(self) -> typing.Tuple[introspection.Pre, ...]:
        """Returns pre-condition validators.

        It is used in the process of generating hypothesis strategies
        To let hypothesis more effectively avoid wrong input values.
        """
        validators = []
        for obj in self._contracts:
            if isinstance(obj, introspection.Pre):
                validators.append(obj)
        return tuple(validators)

    @cached_property
    def exceptions(self) -> typing.Tuple[typing.Type[Exception], ...]:
        """
        Returns exceptions that will be suppressed by individual test cases.
        The exceptions are extracted from `@deal.raises` of the tested function.
        """
        exceptions: list = []
        for obj in self._contracts:
            if isinstance(obj, introspection.Raises):
                exceptions.extend(obj.exceptions)
        return tuple(exceptions)

    @cached_property
    def strategy(self) -> hypothesis.strategies.SearchStrategy:
        """Hypothesis strategy that is used to generate test cases.
        """
        kwargs = self.kwargs.copy()
        for name, value in kwargs.items():
            if isinstance(value, hypothesis.strategies.SearchStrategy):
                continue
            kwargs[name] = hypothesis.strategies.just(value)

        def pass_along_variables(*args, **kwargs) -> ArgsKwargsType:
            return args, kwargs

        pass_along_variables.__signature__ = signature(self.func)    # type: ignore
        update_wrapper(wrapper=pass_along_variables, wrapped=self.func)
        return hypothesis.strategies.builds(pass_along_variables, **kwargs)

    @property
    def _default_settings(self) -> hypothesis.settings:
        return hypothesis.settings(
            database=None,
            max_examples=self.count,
            # avoid showing deal guts
            verbosity=hypothesis.Verbosity.quiet,
            # raise the original exception instead of a fake one
            report_multiple_bugs=False,
            # print how to reproduce the failure
            print_blob=True,
            # if too many cases rejected, it is deal to blame
            suppress_health_check=[hypothesis.HealthCheck.filter_too_much],
        )

    @typing.overload
    def __call__(self, test_func: F) -> F:
        """Wrap a function to turn it into a proper Hypothesis test.

        This is the recommend way to use `deal.cases`. It is powerful and extendable.

        ```python
        >>> import deal
        >>> @deal.pre(lambda a, b: b != 0)
        ... def div(a: int, b: int) -> float:
        ...   return a / b
        ...
        >>> @deal.cases(div)
        ... def test_div(case):
        ...   ...     # do something before
        ...   case()  # run the test case
        ...   ...     # do something after
        ...
        >>> test_div()  # run all test cases for `div`
        >>>
        ```

        """

    @typing.overload
    def __call__(self) -> None:
        """Generate and run tests for a function.

        This is the fastest way to generate tests for a function.

        ```python
        >>> import deal
        >>> @deal.pre(lambda a, b: b != 0)
        ... def div(a: int, b: int) -> float:
        ...   return a / b
        ...
        >>> test_div = deal.cases(div)
        >>> test_div()  # run the test
        ```

        """

    @typing.overload
    def __call__(self, buffer: FuzzInputType) -> typing.Optional[bytes]:
        """Use a function as a fuzzing target.

        This is a way to provide a random buffer for Hypothesis.
        It can be helpful for heavy testing of something really critical.

        ```python
        >>> import deal
        >>> @deal.pre(lambda a, b: b != 0)
        ... def div(a: int, b: int) -> float:
        ...   return a / b
        ...
        >>> import atheris
        >>> test_div = deal.cases(div)
        >>> atheris.Setup([], test_div)
        ...
        >>> atheris.Fuzz()
        ...
        ```

        """

    def __call__(self, target=None):
        """Allows deal.cases to be used as decorator, test function, or fuzzing target.
        """
        __tracebackhide__ = True
        if target is None:
            self._run()
            return None
        if callable(target):
            return self._wrap(target)
        return self._run.hypothesis.fuzz_one_input(target)  # type: ignore[attr-defined]

    # a hack to make the test discoverable by pytest
    @property
    def __func__(self) -> F:
        return self._run

    @cached_property
    def _run(self) -> F:
        return self._wrap(lambda case: case())

    def _wrap(self, test_func: F) -> F:
        # precache all contracts, so hypothesis won't explode
        # because of inconsistent execution time.
        introspection.init_all(test_func)

        def run_examples(args: tuple, kwargs: dict) -> None:
            case = self._make_case()
            for contract in self._contracts:
                if not isinstance(contract, introspection.Example):
                    continue
                case = case._replace(func=contract.validate)
                test_func(case, *args, **kwargs)

        def wrapper(case: ArgsKwargsType, *args, **kwargs) -> None:
            __tracebackhide__ = True
            ex = case
            if ex is EXAMPLE:
                run_examples(args, kwargs)
                return
            for validator in self._pres:
                try:
                    validator.validate(*ex[0], **ex[1])
                except validator.exception_type:
                    hypothesis.reject()
            case = self._make_case(*ex[0], **ex[1])
            test_func(case, *args, **kwargs)

        wrapper = self._impersonate(wrapper=wrapper, wrapped=test_func)
        wrapper = hypothesis.example(case=EXAMPLE)(wrapper)
        wrapper = hypothesis.given(case=self.strategy)(wrapper)
        wrapper = self.settings(wrapper)
        if self.seed is not None:
            wrapper = hypothesis.seed(self.seed)(wrapper)
        return wrapper

    @staticmethod
    def _impersonate(wrapper: F, wrapped: F) -> F:
        if not hasattr(wrapped, '__code__'):
            def wrapped(case) -> None:
                pass
        wrapper = proxies(wrapped)(wrapper)
        if wrapper.__name__ == '<lambda>':
            wrapper.__name__ = 'test_func'
        return wrapper



import ast
from typing import Iterator, List, NamedTuple, Optional, Union

import astroid

from .common import TOKENS, get_name


SUPPORTED_CONTRACTS = frozenset({
    'deal.ensure',
    'deal.example',
    'deal.has',
    'deal.post',
    'deal.pre',
    'deal.pure',
    'deal.raises',
    'deal.safe',
})
SUPPORTED_MARKERS = frozenset({'deal.pure', 'deal.safe', 'deal.inherit'})
Attr = Union[ast.Attribute, astroid.Attribute]


class ContractInfo(NamedTuple):
    name: str
    args: List[Union[ast.expr, astroid.Expr]]
    line: int


def get_contracts(func) -> Iterator[ContractInfo]:
    if isinstance(func, ast.FunctionDef):
        yield from _get_contracts(func.decorator_list)
        return

    assert isinstance(func, (astroid.FunctionDef, astroid.UnboundMethod))
    if func.decorators is None:
        return
    assert isinstance(func.decorators, astroid.Decorators)
    yield from _get_contracts(func.decorators.nodes)


def _get_contracts(decorators: list) -> Iterator[ContractInfo]:
    for contract in decorators:
        if isinstance(contract, TOKENS.ATTR):
            name = get_name(contract)
            if name not in SUPPORTED_MARKERS:
                continue
            yield ContractInfo(
                name=name.split('.')[-1],
                args=[],
                line=contract.lineno,
            )
            if name == 'deal.inherit':
                yield from _resolve_inherit(contract)

        if isinstance(contract, TOKENS.CALL):
            if not isinstance(contract.func, TOKENS.ATTR):
                continue
            name = get_name(contract.func)
            if name == 'deal.chain':
                yield from _get_contracts(contract.args)
            if name not in SUPPORTED_CONTRACTS:
                continue
            yield ContractInfo(
                name=name.split('.')[-1],
                args=contract.args,
                line=contract.lineno,
            )

        # infer assigned value
        if isinstance(contract, astroid.Name):
            assigments = contract.lookup(contract.name)[1]
            if not assigments:
                continue
            # use only the closest assignment
            expr = assigments[0]
            # can it be not an assignment? IDK
            if not isinstance(expr, astroid.AssignName):  # pragma: no cover
                continue
            expr = expr.parent
            if not isinstance(expr, astroid.Assign):  # pragma: no cover
                continue
            yield from _get_contracts([expr.value])


def _resolve_inherit(contract: Attr) -> Iterator[ContractInfo]:
    if not isinstance(contract, astroid.Attribute):
        return
    cls = _get_parent_class(contract)
    if cls is None:
        return
    func = _get_parent_func(contract)
    for base_class in cls.ancestors():
        assert isinstance(base_class, astroid.ClassDef)
        for method in base_class.mymethods():
            assert isinstance(method, astroid.FunctionDef)
            if method.name != func.name:
                continue
            yield from get_contracts(method)


def _get_parent_class(node) -> Optional[astroid.ClassDef]:
    if isinstance(node, astroid.ClassDef):
        return node
    if isinstance(node, (astroid.Attribute, astroid.FunctionDef, astroid.Decorators)):
        return _get_parent_class(node.parent)
    return None


def _get_parent_func(node) -> astroid.FunctionDef:
    if isinstance(node, (astroid.Attribute, astroid.Decorators)):
        return _get_parent_func(node.parent)
    assert isinstance(node, astroid.FunctionDef)
    return node



import deal


def preserve_function_annotation() -> None:
    @deal.inv(lambda obj: obj.likes >= 0)
    class Video:
        likes: int

    v = Video()
    reveal_type(v)  # R: types_inv.Video@6


def infer_validator_type() -> None:
    # E: "Video" has no attribute "hello"
    @deal.inv(lambda obj: obj.hello)
    class Video:
        likes: int



from argparse import ArgumentParser
from contextlib import suppress
from importlib import import_module
from pathlib import Path
from typing import Dict, Iterable, TextIO

from .._colors import COLORS
from .._mem_test import MemoryTracker
from .._state import state
from .._testing import TestCase, cases
from ..linter._extractors.pre import format_call_args
from ._base import Command
from ._common import get_paths
from ._test import get_func_names, sys_path


def run_cases(
    cases: Iterable[TestCase],
    func_name: str,
    stream: TextIO,
    colors: Dict[str, str],
) -> bool:
    print('  {blue}running {name}{end}'.format(name=func_name, **colors), file=stream)
    for case in cases:
        tracker = MemoryTracker()
        debug = state.debug
        state.disable()
        try:
            with tracker, suppress(Exception):
                case()
        finally:
            state.debug = debug
        if not tracker.diff:
            continue

        # show the diff and stop testing the func
        line = '    {yellow}{name}({args}){end}'.format(
            name=func_name,
            args=format_call_args(args=case.args, kwargs=case.kwargs),
            **colors,
        )
        print(line, file=stream)
        longest_name_len = max(len(name) for name in tracker.diff)
        for name, count in tracker.diff.items():
            line = '      {red}{name}{end} x{count}'.format(
                name=name.ljust(longest_name_len),
                count=count,
                **colors,
            )
            print(line, file=stream)
        return False
    return True


class MemtestCommand(Command):
    """Generate and run tests against pure functions and report memory leaks.

    ```bash
    python3 -m deal memtest project/
    ```

    Function must be decorated by one of the following to be run:

    + `@deal.pure`
    + `@deal.has()` (without arguments)

    Options:

    + `--count`: how many input values combinations should be checked.

    Exit code is equal to count of leaked functions.
    See [memory leaks][leaks] documentation for more details.

    [leaks]: https://deal.readthedocs.io/details/tests.html#memory-leaks
    """

    @staticmethod
    def init_parser(parser: ArgumentParser) -> None:
        parser.add_argument('--count', type=int, default=50)
        parser.add_argument('paths', nargs='+')

    def __call__(self, args) -> int:
        failed = 0
        for arg in args.paths:
            for path in get_paths(Path(arg)):
                failed += self.run_tests(
                    path=Path(path),
                    count=args.count,
                )
        return failed

    def run_tests(self, path: Path, count: int) -> int:
        names = list(get_func_names(path=path))
        if not names:
            return 0
        self.print('{magenta}running {path}{end}'.format(path=path, **COLORS))
        module_name = '.'.join(path.relative_to(self.root).with_suffix('').parts)
        with sys_path(path=self.root):
            module = import_module(module_name)
        failed = 0
        for func_name in names:
            func = getattr(module, func_name)
            ok = run_cases(
                cases=cases(func=func, count=count, check_types=False),
                func_name=func_name,
                stream=self.stream,
                colors=COLORS,
            )
            if not ok:
                failed += 1
        return failed



import ast
from typing import Dict, Union

import astroid


TreeType = Union[ast.Module, astroid.Module]
DefsType = Dict[str, ast.stmt]


def get_definitions(tree: TreeType) -> DefsType:
    if isinstance(tree, ast.Module):
        return _extract_defs_ast(tree)
    return _extract_defs_astroid(tree)


def _extract_defs_ast(tree: ast.Module) -> DefsType:
    result: DefsType = dict()
    for node in tree.body:
        if isinstance(node, ast.Import):
            for name_node in node.names:
                name = name_node.asname or name_node.name
                result[name] = ast.Import(
                    names=[name_node],
                    lineno=1,
                    col_offset=1,
                    ctx=ast.Load(),
                )
            continue

        if isinstance(node, ast.ImportFrom):
            if not node.module or node.level:
                continue
            for name_node in node.names:
                name = name_node.asname or name_node.name
                result[name] = ast.ImportFrom(
                    module=node.module,
                    names=[name_node],
                    lineno=1,
                    col_offset=1,
                    ctx=ast.Load(),
                )
            continue

        if isinstance(node, ast.Assign):
            for target in node.targets:
                if not isinstance(target, ast.Name):
                    continue
                result[target.id] = node
    return result


def _extract_defs_astroid(tree: astroid.Module) -> DefsType:
    result: DefsType = dict()
    for node in tree.body:
        if isinstance(node, astroid.Import):
            for name, alias in node.names:
                result[alias or name] = ast.Import(
                    names=[ast.alias(name=name, asname=alias)],
                    lineno=1,
                    col_offset=1,
                    ctx=ast.Load(),
                )
            continue

        if isinstance(node, astroid.ImportFrom):
            if not node.modname or node.level:
                continue
            for name, alias in node.names:
                result[alias or name] = ast.ImportFrom(
                    module=node.modname,
                    names=[ast.alias(name=name, asname=alias)],
                    lineno=1,
                    col_offset=1,
                    ctx=ast.Load(),
                )
            continue

        if isinstance(node, astroid.Assign):
            expr = ast.parse(node.as_string()).body[0]
            for target in node.targets:
                if not isinstance(target, astroid.AssignName):
                    continue
                result[target.name] = expr
    return result



# autopep8: off
import sys
from pathlib import Path

import marshmallow
import pytest
import vaa

import deal
from deal._exceptions import _excepthook, exception_hook
from deal._state import state


def test_source_get_lambda_from_dec():
    @deal.pre(lambda x: x > 0)
    def f(x):
        pass

    with pytest.raises(deal.ContractError) as exc_info:
        f(-2)
    assert exc_info.value.source == 'x > 0'

    state.color = False
    assert str(exc_info.value) == 'expected x > 0 (where x=-2)'
    state.color = True


def test_source_get_lambda_from_dec_simple():
    @deal.pre(lambda _: _.x > 0)
    def f(x):
        pass

    with pytest.raises(deal.ContractError) as exc_info:
        f(-2)
    assert exc_info.value.source == 'x > 0'

    state.color = False
    assert str(exc_info.value) == 'expected x > 0 (where x=-2)'
    state.color = True


def test_source_get_lambda_from_var():
    c = lambda x: x > 0 # noqa

    @deal.pre(c)
    def f(x):
        pass

    with pytest.raises(deal.ContractError) as exc_info:
        f(-2)
    assert exc_info.value.source == 'x > 0'

    state.color = False
    assert str(exc_info.value) == 'expected x > 0 (where x=-2)'
    state.color = True


def test_source_get_lambda_with_braces():
    @deal.pre(lambda x: (x + 1) > (0 + 1))
    def f(x):
        pass

    with pytest.raises(deal.ContractError) as exc_info:
        f(-2)
    assert exc_info.value.source == '(x + 1) > (0 + 1)'


def test_source_get_lambda_multiline_dec():
    @deal.pre(
        lambda x: x > 0,
    )
    def f(x):
        pass

    with pytest.raises(deal.ContractError) as exc_info:
        f(-2)
    assert exc_info.value.source == 'x > 0'


def test_source_get_lambda_multiline_splitted_dec():
    @deal.pre(lambda x: x > 0 and   # noqa
                        x < 10)     # noqa
    def f(x):
        pass

    with pytest.raises(deal.ContractError) as exc_info:
        f(-2)
    assert exc_info.value.source == '<lambda>'


def test_source_get_lambda_from_many():
    @deal.pre(lambda x: x > -10)
    @deal.pre(lambda x: x > 0)
    @deal.pre(lambda x: x > -20)
    def f(x):
        pass

    with pytest.raises(deal.ContractError) as exc_info:
        f(-2)
    assert exc_info.value.source == 'x > 0'


def test_source_get_chain():
    sum_contract = deal.chain(
        deal.pre(lambda a, b: a > 0),
        deal.pre(lambda a, b: b > 0),
        deal.post(lambda res: res > 0),
    )

    @sum_contract
    def sum(a, b):
        return a + b

    with pytest.raises(deal.ContractError) as exc_info:
        sum(-2, 3)
    assert exc_info.value.source == 'a > 0'

    with pytest.raises(deal.ContractError) as exc_info:
        sum(2, -3)
    assert exc_info.value.source == 'b > 0'

    state.color = False
    assert str(exc_info.value) == 'expected b > 0 (where a=2, b=-3)'
    state.color = True


def test_source_get_func_name():
    def identity_contract(x):
        a = 0
        return x > a

    @deal.pre(identity_contract)
    def identity(x):
        return x

    with pytest.raises(deal.ContractError) as exc_info:
        identity(-2)
    assert exc_info.value.source == 'identity_contract'


def test_variables_too_long_repr():
    @deal.pre(lambda a, b: a == b)
    def f(a, b):
        return a + b

    with pytest.raises(deal.ContractError) as exc_info:
        f(234, 'x' * 60)
    state.color = False
    assert exc_info.value.variables == 'a=234'
    state.color = True


def test_source_vaa_scheme():
    class MarshMallowScheme(marshmallow.Schema):
        x = marshmallow.fields.Str()

    @deal.pre(vaa.marshmallow(MarshMallowScheme))
    def identity(x):
        return x

    with pytest.raises(deal.ContractError) as exc_info:
        identity(-2)
    assert exc_info.value.source.startswith('<vaa._external.Marshmallow object at ')
    exp = "[Error(message='Not a valid string.', field='x')] (where x=-2)"
    state.color = False
    assert str(exc_info.value) == exp
    state.color = True


def test_repr_raises_exc():
    @deal.raises()
    def identity(b):
        return 1 / b

    with pytest.raises(deal.RaisesContractError) as exc_info:
        identity(0)
    assert str(exc_info.value) == ''


def test_exception_hook(capsys):
    pre_path = str(Path('deal', '_runtime', '_contracts.py'))
    f = deal.pre(lambda x: x > 0)(lambda x: x)
    with pytest.raises(deal.PreContractError) as exc_info:
        f(-2)

    # custom hook is registered
    assert sys.excepthook is exception_hook

    # the default hook shows the full traceback
    _excepthook(exc_info.type, exc_info.value, exc_info.tb)
    captured = capsys.readouterr()
    assert captured.out == ''
    assert pre_path in captured.err

    # the custom hook hides deal from the traceback
    exception_hook(exc_info.type, exc_info.value, exc_info.tb)
    captured = capsys.readouterr()
    assert captured.out == ''
    # the hook doesn't work on python 3.6 and below
    if sys.version_info[:2] >= (3, 7):
        assert pre_path not in captured.err


def test_exception_hook_ignores_non_contract_exceptions(capsys):
    with pytest.raises(deal.NoMatchError) as exc_info:
        deal.dispatch(lambda: 0)()

    # the custom hook does not reduce non-contract tracebacks
    exception_hook(exc_info.type, exc_info.value, exc_info.tb)
    captured = capsys.readouterr()
    assert captured.out == ''
    base_path = str(Path('deal', '_runtime', '_dispatch.py'))
    assert base_path in captured.err


def test_exception_hook_ignores_contract_from_non_deal(capsys):
    with pytest.raises(deal.ContractError) as exc_info:
        raise deal.ContractError

    # the custom hook does not reduce contract from the traceback
    exception_hook(exc_info.type, exc_info.value, exc_info.tb)
    captured = capsys.readouterr()
    assert captured.out == ''
    assert 'test_exceptions.py' in captured.err


def test_custom_exc():
    @deal.pre(lambda x: x > 0, exception=ZeroDivisionError)
    def f(x):
        pass

    with pytest.raises(ZeroDivisionError) as exc_info:
        f(-2)
    assert exc_info.value.args == ()


def test_custom_exc_with_message():
    @deal.pre(lambda x: x > 0, exception=ZeroDivisionError('oh hi mark'))
    def f(x):
        pass

    with pytest.raises(ZeroDivisionError) as exc_info:
        f(-2)
    assert exc_info.value.args == ('oh hi mark',)


def test_custom_exc_and_message():
    @deal.pre(lambda x: x > 0, exception=ZeroDivisionError, message='oh hi mark')
    def f(x):
        pass

    with pytest.raises(ZeroDivisionError) as exc_info:
        f(-2)
    assert exc_info.value.args == ('oh hi mark',)


def test_custom_exc_and_returned_message():
    @deal.pre(lambda x: x > 0 or 'oh hi mark', exception=ZeroDivisionError)
    def f(x):
        pass

    with pytest.raises(ZeroDivisionError) as exc_info:
        f(-2)
    assert exc_info.value.args == ('oh hi mark',)


def test_vaa_scheme_and_custom_exception():
    @vaa.marshmallow
    class MarshMallowScheme(marshmallow.Schema):
        x = marshmallow.fields.Str()

    @deal.pre(MarshMallowScheme, exception=ZeroDivisionError)
    def identity(x):
        return x

    with pytest.raises(ZeroDivisionError) as exc_info:
        identity(-2)
    exp = "[Error(message='Not a valid string.', field='x')]"
    assert str(exc_info.value) == exp



"""
Get help for libFuzzer:
    python3 examples/fuzzing_atheris.py -help=1

Run 1000 test cases:
    python3 examples/fuzzing_atheris.py -runs=1000
"""
import codecs
import sys

import atheris

import deal


def encode(text: str) -> str:
    return codecs.encode(text, encoding='rot13')


@deal.ensure(lambda text, result: encode(result) == text)
def decode(text: str) -> str:
    assert text != 'bad'
    return codecs.encode(text, encoding='rot13')


def fuzz():
    test = deal.cases(decode)
    atheris.Setup(sys.argv, test)
    atheris.Fuzz()


if __name__ == '__main__':
    fuzz()



import json
from importlib import import_module
from pathlib import Path

import pytest

from deal.linter._contract import Category
from deal.linter._stub import StubFile, StubsManager, _get_funcs, generate_stub


def test_generate_stub_exceptions(tmp_path: Path):
    root = tmp_path / 'project'
    root.mkdir()
    (root / '__init__.py').touch()
    source_path = (root / 'example.py')
    source_path.write_text('def func(): 1/0')
    stub_path = generate_stub(path=source_path)
    content = json.loads(stub_path.read_text())
    assert stub_path.name == 'example.json'
    assert stub_path.parent == root
    assert content == {'func': {'raises': ['ZeroDivisionError']}}


def test_generate_stub_markers(tmp_path: Path):
    root = tmp_path / 'project'
    root.mkdir()
    (root / '__init__.py').touch()
    source_path = (root / 'example.py')
    source_path.write_text('def func(): print("oh hi mark")')
    stub_path = generate_stub(path=source_path)
    content = json.loads(stub_path.read_text())
    assert stub_path.name == 'example.json'
    assert stub_path.parent == root
    assert content == {'func': {'has': ['stdout']}}


def test_generate_stub_bad_ext(tmp_path: Path):
    path = tmp_path / 'example.com'
    with pytest.raises(ValueError, match='invalid.* file extension.*'):
        generate_stub(path=path)


def test_do_not_dump_empty_stub(tmp_path: Path):
    root = tmp_path / 'project'
    root.mkdir()
    (root / '__init__.py').touch()
    source_path = (root / 'example.py')
    source_path.write_text('def func(): return 1')
    stub_path = generate_stub(path=source_path)
    assert not stub_path.exists()
    assert stub_path.name == 'example.json'
    assert stub_path.parent == root


def test_stub_file(tmp_path: Path):
    path = tmp_path / 'example.json'
    stub = StubFile(path=path)

    # add
    stub.add(func='fname', contract=Category.RAISES, value='TypeError')
    with pytest.raises(ValueError, match='unsupported contract'):
        stub.add(func='fname', contract=Category.POST, value='SyntaxError')
    assert stub._content == {'fname': {'raises': ['TypeError']}}

    # do not add twice
    stub.add(func='fname', contract=Category.RAISES, value='TypeError')
    assert stub._content == {'fname': {'raises': ['TypeError']}}

    # get
    assert stub.get(func='fname', contract=Category.RAISES) == frozenset({'TypeError'})
    with pytest.raises(ValueError, match='unsupported contract'):
        stub.get(func='fname', contract=Category.POST)
    assert stub.get(func='unknown', contract=Category.RAISES) == frozenset()

    # dump
    stub.dump()
    content = json.loads(path.read_text(encoding='utf8'))
    assert content == {'fname': {'raises': ['TypeError']}}

    # load
    stub2 = StubFile(path=path)
    stub2.load()
    assert stub2._content == {'fname': {'raises': ['TypeError']}}


@pytest.mark.parametrize('given, expected', [
    ('def f(): pass', ['f']),
    ('async def f(): pass', ['f']),
    ('def f(): pass\n\ndef g(): pass', ['f', 'g']),
    ('class C:\n def f(): pass', ['C.f']),
    ('class A:\n class B:\n  def f(): pass', ['A.B.f']),
    ('nothing\n1\na = 3', []),
])
def test_get_funcs(tmp_path: Path, given: str, expected):
    path = tmp_path / 'example.py'
    path.write_text(given)
    names = [f.name for f in _get_funcs(path=path)]
    assert names == expected


def test_get_module_name(tmp_path: Path):
    root = tmp_path / 'project'
    root.mkdir()
    path = root / 'example.py'
    path.touch()
    assert StubsManager._get_module_name(path=path) == 'example'

    (root / '__init__.py').touch()
    assert StubsManager._get_module_name(path=path) == 'project.example'


@pytest.mark.parametrize('given, expected', [
    ('deal.linter', 'deal.linter.__init__'),
    ('deal._state', 'deal._state'),
    ('pytest', 'pytest.__init__'),
    ('json', 'json.__init__'),
    ('typing', 'typing'),
])
def test_get_module_name_for_real_modules(tmp_path: Path, given, expected):
    module = import_module(given)
    path = Path(module.__file__)
    assert StubsManager._get_module_name(path=path) == expected


def test_stubs_manager(tmp_path: Path):
    stubs = StubsManager()
    root = tmp_path / 'project'
    root.mkdir()
    path = root / 'example.py'

    # test create
    stubs.create(path)
    assert set(stubs._modules) == {'example'}
    assert stubs._modules['example']._content == {}

    # test get
    assert stubs.get('example') is stubs._modules['example']
    expected = {'raises': ['AssertionError', 'TypeError']}
    stub = stubs.get('typing')
    assert stub
    assert stub._content['get_type_hints'] == expected

    # test do not re-create already cached stub
    old_stub = stubs.get('example')
    assert old_stub
    old_stub.dump()
    new_stub = stubs.create(path)
    assert new_stub is old_stub
    assert stubs.get('example') is old_stub

    # the same but path leads to stub, not code
    new_stub = stubs.create(path.with_suffix('.json'))
    assert new_stub is old_stub
    assert stubs.get('example') is old_stub

    # read already dumped stub instead of creating
    old_stub.add(func='fname', contract=Category.RAISES, value='TypeError')
    old_stub.dump()
    stubs = StubsManager()
    new_stub = stubs.create(path)
    assert new_stub is not old_stub
    assert new_stub._content == {'fname': {'raises': ['TypeError']}}

    # test read with non-stub extensions
    stub = stubs.read(path=path)
    assert stub.path.name == 'example.json'
    with pytest.raises(ValueError, match='invalid stub file extension.*'):
        stubs.read(path=path.with_suffix('.cpp'))


def test_marshmallow_get_stubs():
    stubs = StubsManager()
    stub = stubs.get('marshmallow.utils')
    assert stub is not None



import ast
from textwrap import dedent

import astroid
import pytest

from deal.linter._extractors import get_exceptions


@pytest.mark.parametrize('text, expected', [
    ('raise BaseException', (BaseException, )),
    ('raise ValueError', (ValueError, )),
    ('raise UnknownError', ('UnknownError', )),
    ('raise ValueError("lol")', (ValueError, )),
    ('raise unknown()', ()),
    ('raise 1 + 2', ()),
    ('raise', ()),
    ('assert False', (AssertionError, )),
    ('12 / 0', (ZeroDivisionError, )),
    ('12 + 0', ()),
    ('exit()', (SystemExit, )),
    ('exit(13)', (SystemExit, )),
    ('sys.exit(13)', (SystemExit, )),
    ('something.exit(13)', ()),

    # try-except
    ('try:\n raise AError\nexcept Exception:\n pass', ()),
    ('try:\n raise AError\nexcept AError:\n raise BError', ('BError', )),
    ('try:\n pass\nfinally:\n raise KeyError', (KeyError, )),

    ('if True: raise KeyError', (KeyError, )),
    ('for i in lst: raise KeyError', (KeyError, )),
])
def test_get_exceptions_simple(text, expected):
    tree = astroid.parse(text)
    print(tree.repr_tree())
    returns = tuple(r.value for r in get_exceptions(body=tree.body))
    assert returns == expected

    tree = ast.parse(text)
    print(ast.dump(tree))
    returns = tuple(r.value for r in get_exceptions(body=tree.body))
    assert returns == expected


def test_inference_simple():
    text = """
        def subf():
            raise ValueError  # explicit raise

        def subf2():
            c = 1 - 1
            1 / c   # implicit raise
            d = [1, 2, 3]
            1 / d   # resolved into not a constant
            a = b

        @deal.raises(KeyError)
        def f():
            a = 1
            a()         # resolved not in a function
            unknown()   # cannot resolve
            subf()      # resolve
            subf2()
    """
    tree = astroid.parse(dedent(text))
    print(tree.repr_tree())
    func_tree = tree.body[-1].body
    returns = tuple(r.value for r in get_exceptions(body=func_tree))
    assert returns == (ValueError, ZeroDivisionError)


def test_inference_assign():
    text = """
        def subf():
            raise Unknown

        @deal.raises(KeyError)
        def f():
            b = subf()
    """
    tree = astroid.parse(dedent(text))
    print(tree.repr_tree())
    func_tree = tree.body[-1].body
    returns = tuple(r.value for r in get_exceptions(body=func_tree))
    assert returns == ('Unknown', )


def test_inference_ok_uncalled():
    text = """
        def subf():
            raise ValueError

        @deal.raises(KeyError)
        def f():
            subf
    """
    tree = astroid.parse(dedent(text))
    print(tree.repr_tree())
    func_tree = tree.body[-1].body
    returns = tuple(r.value for r in get_exceptions(body=func_tree))
    assert returns == ()


def test_inference_subcalls():
    text = """
        def subf():
            raise ValueError

        def subf2():
            raise IndexError

        @deal.raises(KeyError)
        def f():
            other(subf(), b=subf2())
    """
    tree = astroid.parse(dedent(text))
    print(tree.repr_tree())
    func_tree = tree.body[-1].body
    returns = tuple(r.value for r in get_exceptions(body=func_tree))
    assert returns == (ValueError, IndexError)


def test_resolve_doesnt_fail_for_simple_ast():
    text = """
        def subf():
            raise ValueError  # explicit raise

        @deal.raises(KeyError)
        def f():
            subf()
    """
    tree = ast.parse(dedent(text))
    print(ast.dump(tree))
    func_tree = tree.body[-1].body
    tuple(get_exceptions(body=func_tree))


def test_inference_subcontracts():
    text = """
        @deal.raises(SomeError)     # actual contract
        @deal.raises(1)             # ignore junk
        @deal.post(lambda _: 1)     # ignore other contracts
        def subf():
            return 1

        @deal.raises(KeyError)
        def f():
            b = subf()
    """
    tree = astroid.parse(dedent(text))
    print(tree.repr_tree())
    func_tree = tree.body[-1].body
    returns = tuple(r.value for r in get_exceptions(body=func_tree))
    assert returns == ('SomeError', )


def test_inference_doesnt_have_exceptions():
    text = """
        def subf():
            something()
            return 1

        @deal.raises(KeyError)
        def f():
            b = subf()
    """
    tree = astroid.parse(dedent(text))
    print(tree.repr_tree())
    func_tree = tree.body[-1].body
    returns = tuple(r.value for r in get_exceptions(body=func_tree))
    assert returns == ()



import sys
from argparse import ArgumentParser
from pathlib import Path
from typing import Mapping, Optional, Sequence, TextIO, Type

from ._base import Command


CommandsType = Mapping[str, Type[Command]]


def get_commands() -> CommandsType:
    from ._decorate import DecorateCommand
    from ._lint import LintCommand
    from ._memtest import MemtestCommand
    from ._prove import ProveCommand
    from ._stub import StubCommand
    from ._test import TestCommand

    return dict(
        decorate=DecorateCommand,
        lint=LintCommand,
        memtest=MemtestCommand,
        prove=ProveCommand,
        stub=StubCommand,
        test=TestCommand,
    )


def main(
    argv: Sequence[str], *,
    commands: Optional[CommandsType] = None,
    root: Optional[Path] = None,
    stream: TextIO = sys.stdout,
) -> int:
    if commands is None:
        commands = get_commands()
    if root is None:  # pragma: no cover
        root = Path()
    parser = ArgumentParser(prog='python3 -m deal')
    parser.set_defaults(cmd=None)
    subparsers = parser.add_subparsers()
    for cmd_name, cmd_class in commands.items():
        descr = cmd_class.__doc__ or ''
        descr = (descr.splitlines() or [''])[0]
        subparser = subparsers.add_parser(
            name=cmd_name,
            description=descr,
        )
        cmd = cmd_class(stream=stream, root=root)
        cmd.init_parser(subparser)
        subparser.set_defaults(cmd=cmd)

    try:
        args = parser.parse_args(argv)
    except SystemExit as exc:
        return exc.code
    if args.cmd is None:
        main(['--help'], commands=commands, root=root, stream=stream)
        return 2
    return args.cmd(args)



import ast
from collections import deque
from contextlib import suppress
from functools import partial
from pathlib import Path
from typing import (
    Callable, Dict, Iterator, List, NamedTuple, Optional, Tuple, Type, TypeVar, Union,
)

import astroid

from .._stub import EXTENSION, StubFile, StubsManager


T = TypeVar('T', bound=Type)
N = Tuple[Type[T], Type[T]]
Handler = Callable[..., Union[Optional['Token'], Iterator['Token']]]
Node = Union[ast.AST, astroid.NodeNG]
NodeWithName = Union[astroid.Module, astroid.FunctionDef, astroid.UnboundMethod, astroid.ClassDef]


class TOKENS:
    ASSERT: N[ast.Assert] = (ast.Assert, astroid.Assert)
    ATTR: N[ast.Attribute] = (ast.Attribute, astroid.Attribute)
    BIN_OP: N[ast.BinOp] = (ast.BinOp, astroid.BinOp)
    CALL: N[ast.Call] = (ast.Call, astroid.Call)
    COMPARE: N[ast.Compare] = (ast.Compare, astroid.Compare)
    GLOBAL: N[ast.Global] = (ast.Global, astroid.Global)
    NONLOCAL: N[ast.Nonlocal] = (ast.Nonlocal, astroid.Nonlocal)
    RAISE: N[ast.Raise] = (ast.Raise, astroid.Raise)
    RETURN: N[ast.Return] = (ast.Return, astroid.Return)
    YIELD: N[ast.Yield] = (ast.Yield, astroid.Yield)


DEFAULT_LINE = 0
DEFAULT_COL = 1


class Token(NamedTuple):
    line: int = DEFAULT_LINE
    col: int = DEFAULT_COL
    value: Optional[object] = None
    marker: Optional[str] = None  # marker name or error message


def traverse(body: List[Node]) -> Iterator[Node]:
    for expr in body:
        if isinstance(expr, ast.AST):
            yield from _traverse_ast(expr)
        else:
            yield from _traverse_astroid(expr)


def _traverse_ast(node: ast.AST) -> Iterator[ast.AST]:
    todo = deque([node])
    while todo:
        node = todo.popleft()
        if isinstance(node, ast.Try):
            for h in node.handlers:
                todo.extend(h.body)
            todo.extend(node.orelse)
            todo.extend(node.finalbody)
        else:
            todo.extend(ast.iter_child_nodes(node))
            yield node


def _traverse_astroid(node: astroid.NodeNG) -> Iterator[astroid.NodeNG]:
    todo = deque([node])
    while todo:
        node = todo.popleft()
        if isinstance(node, astroid.TryExcept):
            for h in node.handlers:
                todo.extend(h.body)
            todo.extend(node.orelse)
        else:
            todo.extend(node.get_children())
            yield node


def get_name(expr: Node) -> Optional[str]:
    if isinstance(expr, ast.Name):
        return expr.id
    if isinstance(expr, astroid.Name):
        return expr.name

    if isinstance(expr, astroid.Attribute):
        left = get_name(expr.expr)
        if left is None:
            return None
        return left + '.' + expr.attrname
    if isinstance(expr, ast.Attribute):
        left = get_name(expr.value)
        if left is None:
            return None
        return left + '.' + expr.attr

    return None


def get_full_name(expr: NodeWithName) -> Tuple[str, str]:
    if expr.parent is None:
        return '', expr.name

    if isinstance(expr.parent, astroid.Module):
        return expr.parent.qname(), expr.name

    if isinstance(expr.parent, (astroid.FunctionDef, astroid.UnboundMethod)):
        module_name, func_name = get_full_name(expr=expr.parent)
        return module_name, func_name + '.' + expr.name

    if isinstance(expr.parent, astroid.ClassDef):
        module_name, class_name = get_full_name(expr=expr.parent)
        return module_name, class_name + '.' + expr.name

    path, _, func_name = expr.qname().rpartition('.')
    return path, func_name


def infer(expr: Node) -> Tuple[astroid.NodeNG, ...]:
    if not isinstance(expr, astroid.NodeNG):
        return tuple()
    with suppress(astroid.InferenceError, RecursionError):
        guesses = expr.infer()
        if guesses is astroid.Uninferable:  # pragma: no cover
            return tuple()
        return tuple(g for g in guesses if repr(g) != 'Uninferable')
    return tuple()


def get_stub(
    module_name: Optional[str],
    expr: astroid.FunctionDef,
    stubs: StubsManager,
) -> Optional[StubFile]:
    if not module_name:
        return None
    stub = stubs.get(module_name)
    if stub is not None:
        return stub

    module = _get_module(expr=expr)
    if module is None or module.file is None:
        return None  # pragma: no coverage
    path = Path(module.file).with_suffix(EXTENSION)
    if not path.exists():
        return None
    return stubs.read(path=path)


def _get_module(expr: astroid.NodeNG) -> Optional[astroid.Module]:
    if type(expr) is astroid.Module:
        return expr
    if expr.parent is None:
        return None
    return _get_module(expr.parent)


class Extractor:
    __slots__ = ('handlers', )
    handlers: Dict[type, Handler]

    def __init__(self) -> None:
        self.handlers = dict()

    def register(self, *types: type) -> Callable[[Handler], Handler]:
        return partial(self._register, types)

    def _register(self, types: Tuple[type], handler: Handler) -> Handler:
        for tp in types:
            # it's here to have `getattr` to get nodes from `ast` module
            # that are available only in some Python versions.
            if tp is None:
                continue  # pragma: no coverage
            self.handlers[tp] = handler
        return handler

    def __call__(self, body: List, **kwargs) -> Iterator[Token]:
        for expr in traverse(body=body):
            for token in self._handle(expr=expr, **kwargs):
                yield self._ensure_node_info(expr=expr, token=token)

    def _handle(self, expr: Node, **kwargs) -> Iterator[Token]:
        handler = self.handlers.get(type(expr))
        if handler is None:
            return
        token = handler(expr=expr, **kwargs)
        if token is None:
            return
        if isinstance(token, Token):
            yield token
            return
        yield from token

    @staticmethod
    def _ensure_node_info(token: Token, expr: Union[ast.AST, astroid.NodeNG]) -> Token:
        if token.line == DEFAULT_LINE:
            token = token._replace(line=expr.lineno)
        if token.col == DEFAULT_COL:
            token = token._replace(col=expr.col_offset)
        return token



import inspect
from contextlib import suppress
from functools import lru_cache
from typing import Any, Callable, Dict, Optional, Tuple, Type

import vaa

from .._exceptions import ContractError
from .._types import ExceptionType


@lru_cache(maxsize=16)
def _get_signature(function: Callable) -> inspect.Signature:
    return inspect.signature(function)


def _args_to_vars(
    *,
    args,
    kwargs: Dict[str, Any],
    signature: Optional[inspect.Signature],
    keep_result: bool = True,
) -> Dict[str, Any]:
    """Convert args and kwargs into dict of params based on the given function.

    For simple validators the validator is passed as function.
    """
    if signature is None:
        return kwargs

    params = kwargs.copy()
    # Do not pass argument named `result` into the function.
    # It is a hack for `deal.ensure` with `vaa` validator.
    if not keep_result and 'result' in kwargs:
        kwargs = kwargs.copy()
        del kwargs['result']

    # assign *args to real names
    for name, param in signature.parameters.items():
        params[name] = param.default
    params.update(signature.bind(*args, **kwargs).arguments)
    return params


class Validator:
    __slots__ = (
        'exception',
        'signature',
        'validate',
        'validator',
        'raw_validator',
        'message',
        'function',
    )

    exception: ExceptionType
    signature: Optional[inspect.Signature]
    validate: Any
    validator: Any
    raw_validator: Any
    message: Optional[str]
    function: Any

    def __init__(
        self,
        validator, *,
        message: Optional[str] = None,
        exception: ExceptionType,
    ) -> None:
        self.validate = self._init
        self.raw_validator = validator
        self.message = message
        self.exception = exception
        if message and isinstance(self.exception, type):
            self.exception = self.exception(message)

    @property
    def exception_type(self) -> Type[Exception]:
        if isinstance(self.exception, Exception):
            return type(self.exception)
        return self.exception

    def _make_validator(self) -> Callable:
        """
        If needed, wrap the original raw validator by vaa.
        """
        validator = self.raw_validator
        # implicitly wrap in vaa all external validators
        with suppress(TypeError):
            return vaa.wrap(validator, simple=False)

        # implicitly wrap in vaa.simple only funcs with one `_` argument.
        params = _get_signature(validator).parameters
        if set(params) == {'_'}:
            return vaa.simple(validator, error=self.message)

        return validator

    def _exception(self, *, message: Optional[str] = None, errors=None, params=None) -> Exception:
        exception = self.exception
        if isinstance(exception, Exception):
            if not message and exception.args:
                message = exception.args[0]
            exception = type(exception)

        # if errors provided, use it as error message
        if errors and isinstance(errors, str):
            message = errors
            errors = None

        # raise beautiful ContractError
        if issubclass(exception, ContractError):
            return exception(
                message=message or '',
                validator=self.validator,
                errors=errors,
                params=params,
                origin=getattr(self, 'function', None),
            )

        # raise boring custom exception
        args = []
        if message:
            args.append(message)
        if errors:
            args.append(errors)
        return exception(*args)

    def init(self) -> None:
        self.validator = self._make_validator()
        if hasattr(self.validator, 'is_valid'):
            self.signature = _get_signature(self.function)
            self.validate = self._vaa_validation
        else:
            self.signature = _get_signature(self.validator)
            self.validate = self._simple_validation

    def _init(self, args, kwargs, exc=None) -> None:
        """
        Called as `validator` when the function is called in the first time.
        Does some costly deferred initializations (involving `inspect`).
        Then sets more appropriate validator as `validator` and calls it.
        """
        self.init()
        self.validate(args, kwargs, exc)

    def _vaa_validation(self, args, kwargs, exc=None) -> None:
        """Validate contract using vaa wrapped validator.
        """

        # if it is a decorator for a function, convert positional args into named ones.
        if hasattr(self, 'function'):
            params = _args_to_vars(
                args=args,
                kwargs=kwargs,
                signature=self.signature,
                keep_result=False,
            )
        else:
            params = kwargs

        # validate
        validator = self.validator(data=params)
        if validator.is_valid():
            return

        # if no errors returned, raise the default exception
        errors = validator.errors
        if not errors:
            raise self._exception(params=params) from exc

        # Flatten single error without field to one simple str message.
        # This is for better readability of simple validators.
        if type(errors) is list:  # pragma: no cover
            if type(errors[0]) is vaa.Error:
                if len(errors) == 1:
                    if errors[0].field is None:
                        errors = errors[0].message

        raise self._exception(errors=errors, params=params) from exc

    def _simple_validation(self, args, kwargs, exc=None) -> None:
        """Validate contract using simple validator.

        Simple validator is a validator that reflects the function signature.
        """
        validation_result = self.validator(*args, **kwargs)
        # is invalid (validator returns error message)
        if type(validation_result) is str:
            params = _args_to_vars(args=args, kwargs=kwargs, signature=self.signature)
            raise self._exception(message=validation_result, params=params) from exc
        # is valid (truely result)
        if validation_result:
            return
        # is invalid (falsy result)
        params = _args_to_vars(args=args, kwargs=kwargs, signature=self.signature)
        raise self._exception(params=params) from exc


class RaisesValidator(Validator):
    __slots__ = ('exceptions', )
    exceptions: Tuple[Type[Exception]]

    def __init__(self, exceptions, exception, message) -> None:
        self.exceptions = exceptions
        self.validator = None
        super().__init__(validator=None, message=message, exception=exception)

    def init(self) -> None:
        self.signature = _get_signature(self.function)
        self.validate = self._validate

    def _init(self, args, kwargs, exc=None) -> None:
        self.init()
        self.validate(args, kwargs, exc=exc)

    def _validate(self, args, kwargs, exc=None) -> None:
        exc_type = type(exc)
        if exc_type in self.exceptions:
            return
        # params = _args_to_vars(args=args, kwargs=kwargs, signature=self.signature)
        raise self._exception() from exc_type


class ReasonValidator(Validator):
    __slots__ = ('event', )
    event: Type[Exception]

    def __init__(self, event, **kwargs) -> None:
        self.event = event
        super().__init__(**kwargs)


class InvariantValidator(Validator):
    def init(self) -> None:
        self.signature = None
        self.validator = self._make_validator()
        if hasattr(self.validator, 'is_valid'):
            self.validate = self._vaa_validation
        else:
            self.validate = self._simple_validation

    def _vaa_validation(self, args, kwargs, exc=None) -> None:
        return super()._vaa_validation((), vars(args[0]), exc=exc)



from functools import update_wrapper
from types import FunctionType, MethodType
from typing import Callable, Generic, Optional, TypeVar

from ._contracts import Contracts


F = TypeVar('F', bound=Callable)
ATTR = '__deal_contract'


class Inherit(Generic[F]):
    _func: F
    _func_name: str
    _cls: Optional[type] = None
    __call__: F

    def __init__(self, func: F) -> None:
        self._func = func
        self._func_name = func.__name__
        update_wrapper(wrapper=self, wrapped=func)

    @classmethod
    def wrap(cls, target):
        # wrap function
        if not isinstance(target, type):
            return cls(target)

        # wrap class
        for attr in dir(target):
            func = getattr(target, attr)
            if not isinstance(func, FunctionType):
                continue
            wrapped = cls(func)  # type: ignore[arg-type]
            wrapped.__set_name__(target, attr)
            setattr(target, attr, wrapped)
        return target

    def __set_name__(self, owner: type, name: str) -> None:
        self._cls = owner
        self._func_name = name

    def _patch(self) -> Callable:
        if self._cls is None:
            return self._func

        patched = self._func
        contracts: Optional[Contracts]
        for base in self._cls.mro()[1:]:
            other = getattr(base, self._func.__name__, None)
            if other is None:
                continue
            contracts = getattr(other, ATTR, None)
            if contracts is None:
                continue
            patched = contracts.wrap(patched)

        method = MethodType(patched, self._cls)
        setattr(self._cls, self._func_name, method)
        return method

    def __call__(self, *args, **kwargs):  # type: ignore[no-redef]
        method = self._patch()
        return method(*args, **kwargs)



import pytest

import deal
from deal._imports import deactivate

from .test_runtime.helpers import run_sync


def test_contract_state_switch_custom_param():
    func = deal.pre(lambda x: x > 0)(lambda x: x * 2)
    deal.disable()
    func(-2)
    deal.enable()
    with pytest.raises(deal.PreContractError):
        func(-2)


def test_contract_state_switch_default_param():
    func = deal.pre(lambda x: x > 0)(lambda x: x * 2)
    deal.disable()
    assert func(-2) == -4
    deal.enable()
    with pytest.raises(deal.PreContractError):
        func(-2)


def test_contract_state_switch_default_param_async():
    @deal.pre(lambda x: x > 0)
    async def func(x):
        return x * 2

    deal.disable()
    assert run_sync(func(-2)) == -4
    deal.enable()
    with pytest.raises(deal.PreContractError):
        run_sync(func(-2))


def test_contract_state_switch_default_param_generator():
    @deal.pre(lambda x: x > 0)
    def func(x):
        yield x * 2

    deal.disable()
    assert list(func(-2)) == [-4]
    deal.enable()
    with pytest.raises(deal.PreContractError):
        list(func(-2))


def test_state_switch_module_load():
    with pytest.raises(RuntimeError):
        deal.module_load()
    try:
        deal.disable()
        deal.activate()
        deal.module_load()
    finally:
        deactivate()
        deal.enable()


def test_state_switch_module_load_debug():
    with pytest.raises(RuntimeError):
        deal.module_load()
    try:
        deal.disable()
        deal.activate()
        deal.enable()
    finally:
        deactivate()
        deal.reset()


def test_state_switch_activate():
    try:
        assert deal.activate()
        assert deactivate()

        deal.disable()
        assert not deal.activate()
    finally:
        deactivate()
        deal.enable()



import ast
from typing import Any, Dict, Iterator, Optional, Sequence

import astroid

from .common import Extractor, Token, infer
from .contracts import get_contracts
from .value import UNKNOWN, get_value


get_pre = Extractor()


@get_pre.register(astroid.Call)
def handle_call(expr: astroid.Call, context: Optional[Dict[str, ast.stmt]] = None) -> Iterator[Token]:
    from .._contract import Category, Contract

    args = []
    for subnode in expr.args:
        value = get_value(expr=subnode)
        if value is UNKNOWN:
            return
        args.append(value)

    kwargs: Dict[str, Any] = {}
    for subnode in (expr.keywords or ()):
        value = get_value(expr=subnode.value)
        if value is UNKNOWN:
            return
        kwargs[subnode.arg] = value

    for func in infer(expr.func):
        if not isinstance(func, astroid.FunctionDef):
            continue
        code = f'def f({func.args.as_string()}):0'
        func_args = ast.parse(code).body[0].args  # type: ignore
        for cinfo in get_contracts(func):
            if cinfo.name != 'pre':
                continue
            contract = Contract(
                args=cinfo.args,
                category=Category.PRE,
                func_args=func_args,
                context=context,
            )
            try:
                result = contract.run(*args, **kwargs)
            except NameError:
                continue
            if result is False or type(result) is str:
                yield Token(
                    value=format_call_args(args, kwargs),
                    marker=result or None,
                    line=expr.lineno,
                    col=expr.col_offset,
                )


def format_call_args(args: Sequence, kwargs: Dict[str, Any]) -> str:
    sep = ', '
    args_s = sep.join(map(repr, args))
    items = sorted(kwargs.items())
    kwargs_s = sep.join(f'{k}={repr(v)}' for k, v in items)
    if args and kwargs:
        return args_s + sep + kwargs_s
    return args_s + kwargs_s



"""
The module provides `get_contracts` function which enumerates
contracts wrapping the given function. Every contract is returned
in wrapper providing a stable interface.

Usage example:

```python
import deal

@deal.pre(lambda x: x > 0)
def f(x):
    return x + 1

contracts = deal.introspection.get_contracts(f)
for contract in contracts:
    assert isinstance(contract, deal.introspection.Contract)
    assert isinstance(contract, deal.introspection.Pre)
    assert contract.source == 'x > 0'
    assert contract.exception is deal.PreContractError
    contract.validate(1)
```
"""

from ._extractor import get_contracts, init_all, unwrap
from ._wrappers import (
    Contract, Ensure, Example, Has, Post, Pre, Raises, Reason, ValidatedContract,
)


__all__ = [
    # functions
    'get_contracts',
    'init_all',
    'unwrap',

    # wrappers
    'Contract',
    'Ensure',
    'Example',
    'Has',
    'Post',
    'Pre',
    'Raises',
    'Reason',
    'ValidatedContract',
]



import pytest
import urllib3

import deal

from .helpers import run_sync


def test_raises_expects_function_to_raise_error():
    func = deal.raises(ZeroDivisionError, ValueError)(lambda x: 1 / x)
    with pytest.raises(ZeroDivisionError):
        func(0)
    func(2)

    func = deal.raises(KeyError)(lambda x: 1 / x)
    with pytest.raises(deal.RaisesContractError):
        func(0)


def test_raises_doesnt_override_another_contract():
    @deal.raises(ZeroDivisionError)
    @deal.has()
    def func(do, number):
        if do:
            http = urllib3.PoolManager()
            http.request('GET', 'http://httpbin.org/robots.txt')
        return 1 / number

    assert func(False, 1) == 1
    with pytest.raises(deal.OfflineContractError):
        func(True, 1)
    with pytest.raises(ZeroDivisionError):
        func(False, 0)


def test_decorating_async_function():
    @deal.raises(ZeroDivisionError)
    async def func(x):
        if x == -1:
            raise KeyError
        return 1 / x

    assert run_sync(func(1)) == 1
    with pytest.raises(ZeroDivisionError):
        run_sync(func(0))
    with pytest.raises(deal.RaisesContractError):
        run_sync(func(-1))


def test_raises_doesnt_override_another_contract_async():
    @deal.raises(ZeroDivisionError)
    @deal.has()
    async def func(do, number):
        if do:
            http = urllib3.PoolManager()
            http.request('GET', 'http://httpbin.org/robots.txt')
        return 1 / number

    assert run_sync(func(False, 1)) == 1
    with pytest.raises(deal.OfflineContractError):
        run_sync(func(True, 1))
    with pytest.raises(ZeroDivisionError):
        run_sync(func(False, 0))


def test_decorating_generator():
    @deal.raises(ZeroDivisionError)
    def func(x):
        if x == -1:
            raise KeyError
        yield 1 / x

    assert list(func(1)) == [1]
    with pytest.raises(ZeroDivisionError):
        list(func(0))
    with pytest.raises(deal.RaisesContractError):
        list(func(-1))


def test_raises_generator():
    @deal.raises(ZeroDivisionError)
    @deal.has()
    def func(do, number):
        if do:
            http = urllib3.PoolManager()
            http.request('GET', 'http://httpbin.org/robots.txt')
        yield 1 / number

    assert list(func(False, 1)) == [1]
    with pytest.raises(deal.OfflineContractError):
        list(func(True, 1))
    with pytest.raises(ZeroDivisionError):
        list(func(False, 0))



import ast
from typing import Dict, List, NamedTuple, Optional, Union

import astroid

from .common import TOKENS, get_name
from .value import UNKNOWN, get_value


class Example(NamedTuple):
    args: List[object]
    kwargs: Dict[str, object]
    result: object


def get_example(expr, func_name: str) -> Optional[Example]:
    if not isinstance(expr, TOKENS.COMPARE):
        return None
    if not isinstance(expr.left, TOKENS.CALL):
        return None
    ex = _get_example_from_call(expr.left, func_name)
    if ex is None:
        return None
    result_val = _get_right_value(expr)
    return ex._replace(result=result_val)


def _get_right_value(expr) -> object:
    if isinstance(expr, ast.Compare):
        if len(expr.ops) != 1:
            return UNKNOWN
        if not isinstance(expr.ops[0], (ast.Eq, ast.Is)):
            return UNKNOWN
        return get_value(expr.comparators[0])

    if isinstance(expr, astroid.Compare):
        if len(expr.ops) != 1:
            return UNKNOWN
        op, right = expr.ops[0]
        if op not in {'==', 'is'}:
            return UNKNOWN
        return get_value(right)

    raise RuntimeError('unreachable')


def _get_example_from_call(expr: Union[ast.Call, astroid.Call], func_name: str) -> Optional[Example]:
    if get_name(expr.func) != func_name:
        return None

    args = []
    for arg in expr.args:
        val = get_value(arg)
        if val is UNKNOWN:
            return None
        args.append(val)

    kwargs: Dict[str, object] = {}
    for keyword in (expr.keywords or ()):
        val = get_value(keyword.value)
        if val is UNKNOWN:
            return None
        assert keyword.arg is not None
        kwargs[keyword.arg] = val

    return Example(args, kwargs, None)



"""Entrypoint for mypy plugin.

Activate it in mypy config (pyproject.toml):

```toml
[tool.mypy]
plugins = ["deal.mypy"]
```
"""
# This file is excluded from coverage.

import atexit
import os
from collections import defaultdict
from time import perf_counter
from typing import DefaultDict, List, Optional, Type

from mypy import nodes, types
from mypy.checker import TypeChecker
from mypy.plugin import FunctionSigContext, Plugin


perf: DefaultDict[str, List[float]] = defaultdict(list)
TRACK_PERF = os.getenv('DEAL_TRACK_PERF')


def show_perf() -> None:
    overall = .0
    for name, times in sorted(perf.items()):
        total = sum(times)
        count = len(times)
        mean = total / count
        print(f'{name:30} | {count:5} * {mean:.7f} = {total:.4f}')
        overall += total
    print(f'TOTAL: {overall}')


if TRACK_PERF:
    atexit.register(show_perf)


class DealMypyPlugin(Plugin):
    def __getattribute__(self, name: str):
        if not TRACK_PERF:
            return super().__getattribute__(name)

        start = perf_counter()
        try:
            return super().__getattribute__(name)
        finally:
            now = perf_counter()
            perf[name].append(now - start)

    def get_function_signature_hook(self, fullname: str):
        if fullname == 'deal._runtime._decorators.pre':
            return self._handle_pre
        if fullname == 'deal._runtime._decorators.post':
            return self._handle_post
        if fullname == 'deal._runtime._decorators.ensure':
            return self._handle_ensure
        if fullname == 'deal._runtime._decorators.reason':
            return self._handle_reason
        if fullname == 'deal._runtime._decorators.inv':
            return self._handle_inv
        return None

    def _handle_reason(self, ctx: FunctionSigContext) -> types.CallableType:
        return self._handle_pre(ctx, position=1)

    def _handle_pre(self, ctx: FunctionSigContext, position: int = 0) -> types.CallableType:
        validator = ctx.args[position][0]
        if not isinstance(validator, nodes.LambdaExpr):
            return ctx.default_signature
        if validator.arg_names == ['_']:
            return ctx.default_signature
        dfn = self._get_parent_func(ctx)
        if dfn is None:
            return ctx.default_signature
        ftype = dfn.func.type
        if not isinstance(ftype, types.CallableType):
            return ctx.default_signature
        return self._set_validator_type(
            ctx=ctx,
            ftype=ftype.copy_modified(ret_type=self._val_ret_type(ctx)),
            position=position,
        )

    def _handle_post(self, ctx: FunctionSigContext) -> types.CallableType:
        validator = ctx.args[0][0]
        if not isinstance(validator, nodes.LambdaExpr):
            return ctx.default_signature
        if validator.arg_names == ['_']:
            return ctx.default_signature
        dfn = self._get_parent_func(ctx)
        if dfn is None:
            return ctx.default_signature
        ftype = dfn.func.type
        if not isinstance(ftype, types.CallableType):
            return ctx.default_signature
        return self._set_validator_type(ctx, types.CallableType(
            arg_types=[ftype.ret_type],
            arg_kinds=[nodes.ARG_POS],
            arg_names=[None],
            ret_type=self._val_ret_type(ctx),
            fallback=ftype.fallback,
        ))

    def _handle_ensure(self, ctx: FunctionSigContext) -> types.CallableType:
        validator = ctx.args[0][0]
        if not isinstance(validator, nodes.LambdaExpr):
            return ctx.default_signature
        if validator.arg_names == ['_']:
            return ctx.default_signature
        dfn = self._get_parent_func(ctx)
        if dfn is None:
            return ctx.default_signature
        ftype = dfn.func.type
        if not isinstance(ftype, types.CallableType):
            return ctx.default_signature
        return self._set_validator_type(ctx, types.CallableType(
            arg_types=ftype.arg_types + [ftype.ret_type],
            arg_kinds=ftype.arg_kinds + [nodes.ARG_POS],
            arg_names=ftype.arg_names + ['result'],
            ret_type=self._val_ret_type(ctx),
            fallback=ftype.fallback,
        ))

    def _handle_inv(self, ctx: FunctionSigContext) -> types.CallableType:
        validator = ctx.args[0][0]
        if not isinstance(validator, nodes.LambdaExpr):
            return ctx.default_signature
        if validator.arg_names == ['_']:
            return ctx.default_signature
        dfn = self._get_parent_class(ctx)
        if dfn is None:
            return ctx.default_signature
        ftype = types.CallableType(
            arg_types=[types.Instance(dfn.info, [])],
            arg_kinds=[nodes.ARG_POS],
            arg_names=validator.arg_names,
            ret_type=self._val_ret_type(ctx),
            fallback=types.Instance(dfn.info, []),
        )
        return self._set_validator_type(ctx=ctx, ftype=ftype)

    @staticmethod
    def _set_validator_type(
        ctx: FunctionSigContext,
        ftype: types.CallableType,
        position: int = 0,
    ) -> types.CallableType:
        arg_types = ctx.default_signature.arg_types.copy()
        arg_types[position] = ftype
        return ctx.default_signature.copy_modified(arg_types=arg_types)

    def _get_parent_func(self, ctx: FunctionSigContext) -> Optional[nodes.Decorator]:
        checker = ctx.api
        assert isinstance(checker, TypeChecker)
        return self._find_func(defs=checker.tree.defs, target=ctx.context)

    def _find_func(
        self,
        defs: List[nodes.Statement],
        target: nodes.Context,
    ) -> Optional[nodes.Decorator]:
        for dfn in defs:
            if isinstance(dfn, nodes.Decorator):
                for dec in dfn.decorators:
                    if dec is target:
                        return dfn
                dfn = dfn.func
            if isinstance(dfn, nodes.FuncDef):
                result = self._find_func(defs=dfn.body.body, target=target)
                if result is not None:
                    return result
            if isinstance(dfn, nodes.ClassDef):
                result = self._find_func(defs=dfn.defs.body, target=target)
                if result is not None:
                    return result
        return None

    def _get_parent_class(self, ctx: FunctionSigContext) -> Optional[nodes.ClassDef]:
        checker = ctx.api
        assert isinstance(checker, TypeChecker)
        return self._find_class(defs=checker.tree.defs, target=ctx.context)

    def _find_class(
        self,
        defs: List[nodes.Statement],
        target: nodes.Context,
    ) -> Optional[nodes.ClassDef]:
        for dfn in defs:
            if isinstance(dfn, nodes.ClassDef):
                for dec in dfn.decorators:
                    if dec is target:
                        return dfn
            if isinstance(dfn, nodes.FuncDef):
                result = self._find_class(defs=dfn.body.body, target=target)
                if result is not None:
                    return result
        return None

    def _val_ret_type(self, ctx: FunctionSigContext) -> types.Type:
        return types.UnionType.make_union([
            ctx.api.named_generic_type('builtins.bool', []),
            ctx.api.named_generic_type('builtins.str', []),
        ])


def plugin(version: str) -> Type[Plugin]:
    return DealMypyPlugin



import os
from typing import Callable, TypeVar


T = TypeVar('T', bound=Callable)


class _State:
    __slots__ = ('debug', 'color')
    debug: bool
    color: bool

    def __init__(self) -> None:
        self.reset()

    def reset(self) -> None:
        """Restore contracts switch to default.

        All contracts are disabled on production by default.
        See [runtime][runtime] documentation.

        [runtime]: https://deal.readthedocs.io/basic/runtime.html
        """
        self.debug = __debug__
        self.color = 'NO_COLOR' not in os.environ

    def enable(self) -> None:
        """Enable all contracts.
        """
        self.debug = True

    def disable(self) -> None:
        """Disable all contracts.
        """
        self.debug = False


state = _State()
reset = state.reset
enable = state.enable
disable = state.disable



import ast
from itertools import chain
from typing import List

import astroid

from .common import get_name, traverse


def uses_result(validator) -> bool:
    """Returns True if the validator uses `result` argument.

    It may directly list `result` as the argument,
    or it may be a simple validator and use `_.result` in the body.
    """
    if not isinstance(validator, (ast.Lambda, astroid.Lambda)):
        return True
    if _has_result_arg(validator):
        return True
    if _is_simple_validator(validator):
        return _simple_uses_result(validator)
    return False


def _has_result_arg(validator) -> bool:
    """Returns True if arguments list of the function has `result` argument.
    """
    if isinstance(validator, ast.Lambda):
        args = chain(
            validator.args.args,
            validator.args.kwonlyargs,
        )
        for arg in args:
            if arg.arg == 'result':
                return True
        return False
    if isinstance(validator, astroid.Lambda):
        assert isinstance(validator.args, astroid.Arguments)
        args = chain(
            validator.args.args,
            validator.args.kwonlyargs,
        )
        for arg in args:
            assert isinstance(arg, astroid.AssignName)
            if arg.name == 'result':
                return True
        return False
    raise RuntimeError('unreachable')


def _is_simple_validator(validator) -> bool:
    arg_names: List[str]
    if isinstance(validator, ast.Lambda):
        arg_names = [arg.arg for arg in validator.args.args]
    if isinstance(validator, astroid.Lambda):
        assert isinstance(validator.args, astroid.Arguments)
        arg_names = [arg.name for arg in validator.args.args]
    return arg_names == ['_']


def _simple_uses_result(validator) -> bool:
    """For simple validator, check if `_.result` is used in the body.
    """
    assert isinstance(validator, (ast.Lambda, astroid.Lambda))
    for node in traverse(body=[validator.body]):
        if get_name(node) == '_.result':
            return True
    return False



# This file is excluded from coverage.

from importlib import import_module

import deal
from deal._runtime import Validator


# will be filled from the linter
contract = ...
func = ...


def inject(name: str) -> bool:
    """Import the given module and inject it into globals.
    """
    try:
        globals()[name] = import_module(name)
    except ImportError:
        return False
    return True


def validate(args, kwargs) -> None:
    """Run validator, trying to fix missed imports on the way.
    """
    base = Validator(validator=contract, exception=deal.ContractError)
    if func is not Ellipsis:
        base.function = func

    old_name = None
    for _ in range(10):  # maximum 10 tries, just in case
        try:
            base.validate(args, kwargs)
            return
        except NameError as err:
            # Oh no, we didn't properly inject a variable,
            # and now it is missed. Let's try to import it
            # as module and run the contract again.

            name = err.args[0].split("'")[1]
            # the missed name haven't changed, injection didn't help
            if name == old_name:
                raise
            # try to import missed module
            ok = inject(name)
            # the missed name is not a module, give up
            if not ok:
                raise
            continue


try:
    validate(args, kwargs)  # type: ignore  # noqa: F821
except deal.ContractError as exc:
    result = False
    if exc.args:
        result = exc.args[0]
else:
    result = True



from timeit import Timer
from typing import List


SCALES = (
    (1.0, 's'),
    (1e-3, 'ms'),
    (1e-6, 'us'),
    (1e-9, 'ns'),
)
TICKS = '▁▂▃▄▅▆▇█'
CHUNKS = 6
PRECISION = 3
REPEAT = 5


def format_time(dt):
    for scale, unit in SCALES:
        if dt >= scale:
            break
    value = round(dt / scale, PRECISION)
    return f'{value} {unit}'


def format_amount(number: float, k=0) -> str:
    if number < 1000:
        suffix = 'k' * k
        return f'{number:.0f}{suffix}'
    return format_amount(number / 1000, k + 1)


def run_test(*, name: str, test: str, setup: str = '') -> None:
    print(f'  {name}')
    timer = Timer(stmt=test, setup=setup)
    number, _ = timer.autorange(None)
    if number == 1:
        number, _ = timer.autorange(None)
    raw_timings = timer.repeat(repeat=REPEAT, number=number)
    timings = [dt / number for dt in raw_timings]
    print('    {loops} loops, best of {repeat}: {best} {hist}'.format(
        loops=format_amount(number),
        repeat=REPEAT,
        best=format_time(min(timings)),
        hist=make_histogram(timings),
    ))


def make_histogram(timings: List[float]) -> str:
    """Generate ASCII histogram for the given time points
    """
    best = min(timings)
    worst = max(timings)
    diff = worst - best

    if diff == 0:
        return TICKS[-1] * CHUNKS

    histogram = []
    for time in timings:
        ratio = (time - best) / diff
        index = int(round(ratio * CHUNKS))
        histogram.append(TICKS[index])
    return ''.join(histogram)



from pathlib import Path
from textwrap import dedent

import pytest

from deal.linter import TransformationType, Transformer


@pytest.mark.parametrize('content', [
    # add deal.safe
    """
        def f():
            return 1
        ---
        @deal.safe
        def f():
            return 1
    """,
    # preserve deal.safe
    """
        @deal.safe
        def f():
            return 1
        ---
        @deal.safe
        def f():
            return 1
    """,
    # preserve deal.safe with comments
    """
        @deal.safe   # oh hi mark
        def f():
            return 1
        ---
        @deal.safe   # oh hi mark
        def f():
            return 1
    """,
    # preserve deal.raises
    """
        @deal.raises(KeyError)
        def f():
            return 1
        ---
        @deal.raises(KeyError)
        def f():
            return 1
    """,
    """
        @deal.raises(KeyError, UnknownError)
        def f():
            return 1
        ---
        @deal.raises(KeyError, UnknownError)
        def f():
            return 1
    """,
    # add a new deal.raises
    """
        def f():
            raise ValueError
        ---
        @deal.raises(ValueError)
        def f():
            raise ValueError
    """,
    # add deal.raises for unknown error
    """
        def f():
            raise UnknownError
        ---
        @deal.raises(UnknownError)
        def f():
            raise UnknownError
    """,
    # preserve unknown error
    """
        @deal.raises(UnknownError)
        def f():
            raise ValueError
        ---
        @deal.raises(UnknownError, ValueError)
        def f():
            raise ValueError
    """,
    # remove deal.safe if adding deal.raises
    """
        @deal.safe
        def f():
            raise ValueError
        ---
        @deal.raises(ValueError)
        def f():
            raise ValueError
    """,
    # remove deal.pure if adding deal.raises
    """
        @deal.pure
        def f():
            raise ValueError
        ---
        @deal.raises(ValueError)
        @deal.has()
        def f():
            raise ValueError
    """,
    # merge deal.raises
    """
        @deal.raises(ZeroDivisionError)
        def f():
            raise ValueError
        ---
        @deal.raises(ZeroDivisionError, ValueError)
        def f():
            raise ValueError
    """,
    # preserve comments
    """
        # oh hi mark
        def f():
            raise ValueError
        ---
        # oh hi mark
        @deal.raises(ValueError)
        def f():
            raise ValueError
    """,
    # preserve comments
    """
        # oh hi mark

        def f():
            # hello
            raise ValueError
        ---
        # oh hi mark

        @deal.raises(ValueError)
        def f():
            # hello
            raise ValueError
    """,
    # preserve contracts
    """
        @deal.safe
        @deal.pre(lambda: True)
        def f():
            return 1
        ---
        @deal.safe
        @deal.pre(lambda: True)
        def f():
            return 1
    """,
    """
        @deal.pre(lambda: True)
        def f():
            return 1
        ---
        @deal.safe
        @deal.pre(lambda: True)
        def f():
            return 1
    """,
    """
        @deal.raises(ValueError)
        @deal.pre(lambda: True)
        def f():
            1/0
        ---
        @deal.raises(ValueError, ZeroDivisionError)
        @deal.pre(lambda: True)
        def f():
            1/0
    """,
    # support methods
    """
        class A:
            def f(self):
                1/0
        ---
        class A:
            @deal.raises(ZeroDivisionError)
            def f(self):
                1/0
    """,
    # insert after @property
    """
        class A:
            @property
            def f(self):
                1/0
        ---
        class A:
            @property  # type: ignore[misc]
            @deal.raises(ZeroDivisionError)
            def f(self):
                1/0
    """,
    # do not touch @property if no mutations
    """
        class A:
            @property
            @deal.safe
            def f(self):
                pass
        ---
        class A:
            @property
            @deal.safe
            def f(self):
                pass
    """,
    # insert before staticmethod
    """
        class A:
            @staticmethod
            def f(self):
                1/0
        ---
        class A:
            @deal.raises(ZeroDivisionError)
            @staticmethod
            def f(self):
                1/0
    """,
    # insert before multiline decorators
    """
        class A:
            @unknown(
                hi,
            )
            def f(self):
                1/0
        ---
        class A:
            @deal.raises(ZeroDivisionError)
            @unknown(
                hi,
            )
            def f(self):
                1/0
    """,
])
def test_transformer_raises(content: str, tmp_path: Path) -> None:
    given, expected = content.split('---')
    given = dedent(given)
    expected = dedent(expected)
    tr = Transformer(
        content=given,
        path=tmp_path / 'example.py',
        types={TransformationType.RAISES, TransformationType.SAFE},
    )
    actual = tr.transform()
    print(tr.mutations)
    assert actual == expected


@pytest.mark.parametrize('content', [
    # add deal.has()
    """
        def f():
            return 1
        ---
        @deal.has()
        def f():
            return 1
    """,
    # add deal.has with markers
    """
        def f():
            print("hi")
            return 1
        ---
        @deal.has('stdout')
        def f():
            print("hi")
            return 1
    """,
    """
        import random
        def f():
            print(random.choice([1,2]))
            return 1
        ---
        import random
        @deal.has('random', 'stdout')
        def f():
            print(random.choice([1,2]))
            return 1
    """,
    # merge deal.has
    """
        @deal.has('random')
        def f():
            print("hi")
            return 1
        ---
        @deal.has('random', 'stdout')
        def f():
            print("hi")
            return 1
    """,
    """
        @deal.has()
        def f():
            print("hi")
            return 1
        ---
        @deal.has('stdout')
        def f():
            print("hi")
            return 1
    """,
    # drop invalid markers
    """
        @deal.has(13, None, 'random', [])
        def f():
            print("hi")
            return 1
        ---
        @deal.has('random', 'stdout')
        def f():
            print("hi")
            return 1
    """,
    # replace deal.pure
    """
        @deal.pure
        def f():
            print("hi")
            return 1
        ---
        @deal.has('stdout')
        @deal.safe
        def f():
            print("hi")
            return 1
    """,
    # preserve contracts
    """
        @deal.pre(lambda: True)
        def f():
            print("hi")
            return 1
        ---
        @deal.has('stdout')
        @deal.pre(lambda: True)
        def f():
            print("hi")
            return 1
    """,
    """
        @deal.has('random')
        def f():
            return 1
        ---
        @deal.has('random')
        def f():
            return 1
    """,
    """
        @deal.has('random')
        @deal.pre(lambda: True)
        def f():
            return 1
        ---
        @deal.has('random')
        @deal.pre(lambda: True)
        def f():
            return 1
    """,
    """
        @deal.has()
        @deal.pre(lambda: True)
        def f():
            return 1
        ---
        @deal.has()
        @deal.pre(lambda: True)
        def f():
            return 1
    """,
    """
        @deal.pre(lambda: True)
        def f():
            return 1
        ---
        @deal.has()
        @deal.pre(lambda: True)
        def f():
            return 1
    """,
    # do not add deal.raises if transformation type is disabled
    """
        def f():
            raise ValueError
            return 1
        ---
        @deal.has()
        def f():
            raise ValueError
            return 1
    """,
])
def test_transformer_has(content: str, tmp_path: Path) -> None:
    given, expected = content.split('---')
    given = dedent(given)
    expected = dedent(expected)
    tr = Transformer(
        content=given,
        path=tmp_path / 'example.py',
        types={TransformationType.HAS},
    )
    actual = tr.transform()
    assert actual == expected


@pytest.mark.parametrize('content', [
    # add import if needed
    """
        def f():
            return 1
        ---
        import deal
        @deal.has()
        def f():
            return 1
    """,
    # skip imports
    """
        import re

        def f():
            return 1
        ---
        import re
        import deal

        @deal.has()
        def f():
            return 1
    """,
    # skip import-from, do not skip consts
    """
        import re
        from textwrap import dedent

        HI = 1

        def f():
            return 1
        ---
        import re
        import deal
        from textwrap import dedent

        HI = 1

        @deal.has()
        def f():
            return 1
    """,
    # do nothing if there are no mutations
    """
        @deal.has()
        def f():
            return 1
        ---
        @deal.has()
        def f():
            return 1
    """,
    # do not duplicate existing imports
    """
        import deal
        def f():
            return 1
        ---
        import deal
        @deal.has()
        def f():
            return 1
    """,
    # support multiline imports
    """
        from textwrap import (
            dedent,
        )
        def f():
            return 1
        ---
        import deal
        from textwrap import (
            dedent,
        )
        @deal.has()
        def f():
            return 1
    """,
    # skip __future__ imports
    """
        from __future__ import annotations
        def f():
            return 1
        ---
        from __future__ import annotations
        import deal
        @deal.has()
        def f():
            return 1
    """,
    # skip from imports before module imports
    """
        from __future__ import annotations
        def f():
            return 1
        ---
        from __future__ import annotations
        import deal
        @deal.has()
        def f():
            return 1
    """,
])
def test_transformer_import(content: str, tmp_path: Path) -> None:
    given, expected = content.split('---')
    given = dedent(given).lstrip('\n')
    expected = dedent(expected)
    tr = Transformer(
        content=given,
        path=tmp_path / 'example.py',
        types={TransformationType.HAS, TransformationType.IMPORT},
    )
    actual = tr.transform()
    assert actual.lstrip('\n') == expected.lstrip('\n')



import ast

import astroid
import pytest

from deal.linter._extractors import get_imports


@pytest.mark.parametrize('text, expected', [
    ('from deal import pre', ('deal', )),
])
def test_get_imports_simple(text, expected):
    tree = astroid.parse(text)
    print(tree.repr_tree())
    returns = tuple(r.value for r in get_imports(body=tree.body))
    assert returns == expected

    tree = ast.parse(text)
    print(ast.dump(tree))
    returns = tuple(r.value for r in get_imports(body=tree.body))
    assert returns == expected



from argparse import ArgumentParser
from pathlib import Path
from typing import List

from ..linter import StubsManager, generate_stub
from ._base import Command
from ._common import get_paths


class StubCommand(Command):
    """Generate stub files for the given Python files.

    ```bash
    python3 -m deal stub project/
    ```

    Options:

    + `--iterations`: how many time run stub generation against files.
      Every new iteration uses results from the previous ones, improving the result.
      Default: 1.

    Exit code is 0. See [stubs][stubs] documentation for more details.

    [stubs]: https://deal.readthedocs.io/details/stubs.html
    """

    @staticmethod
    def init_parser(parser: ArgumentParser) -> None:
        parser.add_argument('--iterations', type=int, default=1)
        parser.add_argument('paths', nargs='+')

    def __call__(self, args) -> int:
        paths: List[Path] = []
        for arg in args.paths:
            for path in get_paths(Path(arg)):
                paths.append(path)

        roots = list(StubsManager.default_paths) + list(set(paths))
        stubs = StubsManager(paths=roots)

        for _ in range(args.iterations):
            for path in paths:
                generate_stub(path=path, stubs=stubs)
        return 0



import gc
import typing
from collections import Counter

from ._cached_property import cached_property


class MemoryTracker:
    before: typing.Counter[str]
    after: typing.Counter[str]

    def __init__(self) -> None:
        self.before = Counter()
        self.after = Counter()

    def __enter__(self) -> None:
        self.before = self._dump()

    def __exit__(self, *exc) -> None:
        self.after = self._dump()

    @cached_property
    def diff(self) -> typing.Counter[str]:
        return self.after - self.before - Counter({'weakref': 1})

    @staticmethod
    def _dump() -> typing.Counter[str]:
        counter: typing.Counter[str] = Counter()
        gc.collect()
        for obj in gc.get_objects():
            name: str = type(obj).__qualname__
            counter[name] += 1
        return counter



from typing import FrozenSet, Optional, Tuple, Type

from .._cached_property import cached_property
from .._runtime import HasPatcher, RaisesValidator, ReasonValidator, Validator
from .._source import get_validator_source
from .._types import ExceptionType


class Contract:
    """The base class for all contracts returned by `get_contracts`.
    """
    __slots__ = ('_wrapped',)
    _wrapped: Validator

    def __init__(self, wrapped) -> None:
        self._wrapped = wrapped

    @property
    def exception(self) -> ExceptionType:
        """The exception raised if the contract is not satisfied.
        """
        return self._wrapped.exception

    @property
    def exception_type(self) -> Type[Exception]:
        """The type of the exception raised if the contract is not satisfied.
        """
        return self._wrapped.exception_type

    @property
    def message(self) -> Optional[str]:
        """The message (contract description) provided by user.
        """
        return self._wrapped.message


class ValidatedContract(Contract):
    """The base class for all contracts supporting validation.
    """

    def init(self) -> None:
        """Initialize the contract.

        This method is called when the contract is validated in the first time.
        It includes some costly operations, like introspection of the wrapped function.
        You can call it manually if you want to initialize the contract before validation.
        For instance, if you need a consistent execution time for the function.
        """
        self._wrapped.init()

    def validate(self, *args, **kwargs) -> None:
        """Run the validator.

        If validation fails, the `exception` is raised.
        """
        self._wrapped.validate(args, kwargs)

    @cached_property
    def source(self) -> str:
        """The source code for the validator function.

        For named functions it is the name of the function.
        For lambdas it is the body of the lambda.
        """
        self._wrapped.init()
        return get_validator_source(self._wrapped.validator)


class Pre(ValidatedContract):
    """Wrapper for `deal.pre`.
    """


class Post(ValidatedContract):
    """Wrapper for `deal.post`.
    """


class Ensure(ValidatedContract):
    """Wrapper for `deal.ensure`.
    """


class Example(ValidatedContract):
    """Wrapper for `deal.example`.
    """


class Raises(Contract):
    """Wrapper for `deal.raises`.
    """
    _wrapped: RaisesValidator

    @property
    def exceptions(self) -> Tuple[Type[Exception], ...]:
        """Exceptions that the function can raise.
        """
        return self._wrapped.exceptions


class Reason(ValidatedContract):
    """Wrapper for `deal.reason`.
    """
    _wrapped: ReasonValidator

    @property
    def event(self) -> Type[Exception]:
        """The exception for which the validator is provided.
        """
        return self._wrapped.event


class Has(Contract):
    """Wrapper for `deal.has`.
    """
    __slots__ = ('_patcher',)
    _patcher: HasPatcher

    def __init__(self, wrapped) -> None:
        self._patcher = wrapped

    @property
    def exception(self) -> ExceptionType:
        return self._patcher.exception

    @property
    def exception_type(self) -> Type[Exception]:
        return self._patcher.exception_type

    @property
    def message(self) -> Optional[str]:
        return self._patcher.message

    @property
    def markers(self) -> FrozenSet[str]:
        """Side-effects that the function may have.
        """
        return self._patcher.markers



from ._checker import Checker
from ._stub import StubsManager, generate_stub
from ._transformer import TransformationType, Transformer


__all__ = [
    'Checker',
    'generate_stub',
    'StubsManager',
    'TransformationType',
    'Transformer',
]



import ast
import builtins
import enum
from copy import copy
from pathlib import Path
from typing import Dict, FrozenSet, Iterable, List, Optional, Type, Union

import astroid

from .._cached_property import cached_property


TEMPLATE = (Path(__file__).parent / '_template.py').read_text()
CONTRACT_INDEX = 3
FUNC_INDEX = 4


class Category(enum.Enum):
    ENSURE = 'ensure'
    EXAMPLE = 'example'
    HAS = 'has'
    POST = 'post'
    PRE = 'pre'
    PURE = 'pure'
    RAISES = 'raises'
    SAFE = 'safe'
    INHERIT = 'inherit'

    @property
    def brackets_optional(self) -> bool:
        return self in {Category.SAFE, Category.PURE}


class Contract:
    args: tuple
    category: Category
    func_args: ast.arguments
    context: Dict[str, ast.stmt]
    line: int

    def __init__(
        self,
        args: Iterable,
        category: Category,
        func_args: ast.arguments,
        context: Optional[Dict[str, ast.stmt]] = None,
        line: int = 0,
    ) -> None:
        self.args = tuple(args)
        self.category = category
        self.func_args = func_args
        self.context = context or dict()
        self.line = line

    @cached_property
    def body(self) -> ast.AST:
        contract = self.args[0]
        # convert astroid node to ast node
        if hasattr(contract, 'as_string'):
            contract = self._resolve_name(contract)
            contract = ast.parse(contract.as_string()).body[0]
        return contract

    @cached_property
    def arguments(self) -> FrozenSet[str]:
        """Contract function arguments names.

        Useful for resolving external dependencies.
        """
        func = self.body
        if isinstance(func, ast.Expr):
            func = func.value
        if not isinstance(func, (ast.FunctionDef, ast.Lambda)):
            return frozenset()
        args = func.args
        result = set()
        for arg in args.args:
            result.add(arg.arg)
        for arg in args.kwonlyargs:
            result.add(arg.arg)
        if args.vararg:
            result.add(args.vararg.arg)
        if args.kwarg:
            result.add(args.kwarg.arg)
        return frozenset(result)

    @cached_property
    def dependencies(self) -> FrozenSet[str]:
        """Names that are defined outside of the contract body.

        1. Excludes built-in objects.
        1. Excludes contract function arguments.
        """
        deps = set()
        for node in ast.walk(self.body):
            if not isinstance(node, ast.Name):
                continue
            if hasattr(builtins, node.id):
                continue
            if node.id in self.arguments:
                continue
            deps.add(node.id)
        return frozenset(deps)

    @staticmethod
    def _resolve_name(contract):
        if not isinstance(contract, astroid.Name):
            return contract
        definitions = contract.lookup(contract.name)[1]
        if not definitions:
            return contract
        definition = definitions[0]
        if isinstance(definition, astroid.FunctionDef):
            return definition
        if isinstance(definition, astroid.AssignName):
            return definition.parent.value
        # resolved into something tricky, live with it
        return contract  # pragma: no cover

    @cached_property
    def exceptions(self) -> List[Union[str, Type[Exception]]]:
        from ._extractors import get_name

        excs = []
        for expr in self.args:
            name = get_name(expr)
            if not name:
                continue
            exc = getattr(builtins, name, name)
            excs.append(exc)
        return excs

    @cached_property
    def module(self) -> ast.Module:
        module = ast.parse(TEMPLATE)

        # inject function signature
        func = ast.Lambda(
            args=self.func_args,
            body=ast.Set(
                elts=[],
                lineno=1,
                col_offset=1,
                ctx=ast.Load(),
            ),
            lineno=1,
            col_offset=1,
            ctx=ast.Load(),
        )
        module.body[FUNC_INDEX].value = func  # type: ignore

        # collect definitions for contract external deps
        deps: List[ast.stmt] = []
        for dep in self.dependencies:
            definition = self.context.get(dep)
            if not definition:
                continue
            deps.append(definition)

        # inject contract if contract is a function
        contract = copy(self.body)
        if isinstance(contract, ast.FunctionDef):
            contract.body = deps + contract.body
            # if contract is function, add it's definition and assign it's name
            # to `contract` variable.
            module.body = [contract] + module.body      # type: ignore
            module.body[FUNC_INDEX].value = ast.Name(   # type: ignore
                id=contract.name,
                lineno=1,
                col_offset=1,
                ctx=ast.Load(),
            )
            return module

        if isinstance(contract, ast.Expr):
            contract = contract.value

        # Inject contract if contract is a lambda.
        # We have to rebuild lambda into a function
        # to inject dependencies inside the body.
        if isinstance(contract, ast.Lambda):
            body = list(deps)
            return_node = ast.Return(
                value=contract.body,
                lineno=1,
                col_offset=1,
                ctx=ast.Load(),
            )
            body.append(return_node)
            module.body[CONTRACT_INDEX] = ast.FunctionDef(
                name='contract',
                args=contract.args,
                body=body,
                decorator_list=[],
                lineno=1,
                col_offset=1,
                ctx=ast.Load(),
            )
            return module

        # inject contract if contract is an unknown expression
        module.body[CONTRACT_INDEX].value = contract  # type: ignore
        return module

    @cached_property
    def bytecode(self):
        return compile(self.module, filename='<ast>', mode='exec')

    def run(self, *args, **kwargs):
        globals = dict(args=args, kwargs=kwargs)
        exec(self.bytecode, globals)
        return globals['result']

    def __repr__(self) -> str:
        return f'{type(self).__name__}({self.category.value})'



import pytest

import deal


def test_inherit_one_parent():
    class A:
        @deal.pre(lambda self, x: x == 3)
        def f(self, x):
            raise NotImplementedError

    class B(A):
        @deal.inherit
        def f(self, x):
            return x * 2

    b = B()
    assert b.f(3) == 6
    with pytest.raises(deal.PreContractError):
        b.f(4)


def test_inherit_no_parents():
    class A:
        @deal.inherit
        def f(self, x):
            return x * 2

    a = A()
    assert a.f(3) == 6
    assert a.f(4) == 8


def test_inherit_function():
    @deal.inherit
    def f(x):
        return x * 2

    assert f(3) == 6
    assert f(4) == 8


def test_inherit_one_parent_no_contracts():
    class A:
        def f(self, x):
            raise NotImplementedError

    class B(A):
        @deal.inherit
        def f(self, x):
            return x * 2

    b = B()
    assert b.f(3) == 6
    assert b.f(4) == 8


def test_inherit_multiple_parents():
    class A:
        def f(self, x):
            raise NotImplementedError

    class B:
        @deal.pre(lambda self, x: x > 0)
        def f(self, x):
            raise NotImplementedError

    class C(B):
        @deal.pre(lambda self, x: x < 5)
        def f(self, x):
            raise NotImplementedError

    class D(C, A):
        @deal.inherit
        def f(self, x):
            return x * 2

    d = D()
    assert d.f(3) == 6
    with pytest.raises(deal.PreContractError):
        assert d.f(-2) == -4
    with pytest.raises(deal.PreContractError):
        d.f(8)


def test_inherit_one_parent__preserve_contracts():
    class A:
        @deal.pre(lambda self, x: x < 5)
        def f(self, x):
            raise NotImplementedError

    class B(A):
        @deal.inherit
        @deal.pre(lambda self, x: x > 0)
        def f(self, x):
            return x * 2

    b = B()
    assert b.f(3) == 6
    with pytest.raises(deal.PreContractError):
        b.f(6)
    with pytest.raises(deal.PreContractError):
        b.f(-2)


def test_inherit_class():
    class A:
        @deal.pre(lambda self, x: x == 3)
        def f(self, x):
            raise NotImplementedError

    @deal.inherit
    class B(A):
        def f(self, x):
            return x * 2

    b = B()
    assert b.f(3) == 6
    with pytest.raises(deal.PreContractError):
        b.f(4)


def test_inherit_inherit():
    class A:
        @deal.inherit
        @deal.pre(lambda self, x: x == 3)
        def f(self, x):
            pass

    class B(A):
        @deal.inherit
        def f(self, x):
            return x * 2

    b = B()
    assert b.f(3) == 6
    with pytest.raises(deal.PreContractError):
        b.f(4)


def test_has_inherit():
    class A:
        @deal.has('a', 'b')
        def f(self, x):
            raise NotImplementedError

    @deal.inherit
    class B(A):
        def f(self, x):
            return x * 2

    b = B()
    contract = next(deal.introspection.get_contracts(b.f))
    assert isinstance(contract, deal.introspection.Has)
    assert contract.markers == frozenset({'a', 'b'})


def test_has_inherit_and_merge():
    class A:
        @deal.has('a', 'b')
        def f(self, x):
            raise NotImplementedError

    @deal.inherit
    class B(A):
        @deal.has('c', 'd')
        def f(self, x):
            return x * 2

    b = B()
    contract = next(deal.introspection.get_contracts(b.f))
    assert isinstance(contract, deal.introspection.Has)
    assert contract.markers == frozenset({'a', 'b', 'c', 'd'})


def test_has_preserve():
    class A:
        def f(self, x):
            raise NotImplementedError

    @deal.inherit
    class B(A):
        @deal.has('a', 'b')
        def f(self, x):
            return x * 2

    b = B()
    contract = next(deal.introspection.get_contracts(b.f))
    assert isinstance(contract, deal.introspection.Has)
    assert contract.markers == frozenset({'a', 'b'})



import ast

import astroid

from .common import Extractor, Token


get_imports = Extractor()


@get_imports.register(astroid.ImportFrom)
def handle_astroid(expr: astroid.ImportFrom) -> Token:
    dots = '.' * (expr.level or 0)
    name = expr.modname or ''
    return Token(value=dots + name, line=expr.lineno, col=expr.col_offset)


@get_imports.register(ast.ImportFrom)
def handle_ast(expr: ast.ImportFrom) -> Token:
    dots = '.' * expr.level
    name = expr.module or ''
    return Token(value=dots + name, line=expr.lineno, col=expr.col_offset)




from typing import Dict

import pygments
from pygments.formatters import TerminalFormatter
from pygments.lexers import PythonLexer

from ._state import state


COLORS = dict(
    red='\033[91m',
    green='\033[92m',
    yellow='\033[93m',
    blue='\033[94m',
    magenta='\033[95m',
    end='\033[0m',
)
NOCOLORS = dict(
    red='',
    green='',
    yellow='',
    blue='',
    magenta='',
    end='',
)


def highlight(source: str) -> str:
    source = pygments.highlight(
        code=source,
        lexer=PythonLexer(),
        formatter=TerminalFormatter(),
    )
    return source.rstrip()


def get_colors(args) -> Dict[str, str]:
    if not state.color:
        return NOCOLORS
    if args.nocolor:
        state.color = False
        return NOCOLORS
    return COLORS



import pytest

import deal
from deal._trace import Only, _get_func_body_statements, format_lines, trace


def just_return() -> int:
    return 123


def cond_return(cond=False) -> int:
    if cond:
        return 123
    return 456


@deal.safe
@deal.has()
def cond_return_dec(cond=False) -> int:
    if cond:
        return 123
    return 456


def call_another():
    cond = something_else()
    return cond_return(cond=cond)


def something_else() -> bool:
    return False


def test_trace_100():
    result = trace(just_return)
    assert len(result.covered_lines) == 1
    assert len(result.all_lines) == 1
    assert result.func_result == 123


def test_trace_33():
    result = trace(cond_return)
    assert len(result.covered_lines) == 2
    assert len(result.all_lines) == 3
    assert result.func_result == 456


def test_trace_33_dec():
    result = trace(cond_return_dec)
    assert len(result.covered_lines) == 2
    assert len(result.all_lines) == 3
    assert result.func_result == 456


def test_trace_100_exclude_everythong_else():
    result = trace(call_another)
    assert len(result.covered_lines) == 2
    assert len(result.all_lines) == 2
    assert result.func_result == 456


@pytest.mark.parametrize('statements, lines, result', [
    ({1, 2, 3, 4, 5, 10, 11, 12, 13, 14}, {1, 2, 5, 10, 11, 13, 14}, '1-2, 5-11, 13-14'),
    ({1, 2, 3, 4, 5, 10, 11, 12, 13, 14, 98, 99}, {1, 2, 5, 10, 11, 13, 14, 99}, '1-2, 5-11, 13-14, 99'),
    ({1, 2, 3, 4, 98, 99, 100, 101, 102, 103, 104}, {1, 2, 99, 102, 103, 104}, '1-2, 99, 102-104'),
    ({17}, {17}, '17'),
    ({90, 91, 92, 93, 94, 95}, {90, 91, 92, 93, 94, 95}, '90-95'),
    ({1, 2, 3, 4, 5}, set(), ''),
    ({1, 2, 3, 4, 5}, {4}, '4'),
])
def test_format_lines(statements, lines, result):
    assert format_lines(statements, lines) == result


def big_test_func(cond):
    """docstring
    """
    if cond:
        return dict(
            a=1,
            b=2,
            c=3,
        )
    return 42


def test_get_func_body_statements():
    stmts = _get_func_body_statements(big_test_func)
    assert len(stmts) == 3


def test_get_func_body_statements_no_func():
    stmts = _get_func_body_statements(lambda x: x)
    assert len(stmts) == 1


def test_only():
    only = Only('excl.py')
    assert only.names('excl.py', 'anything') == 0
    assert only.names('something_else.py', 'anything') == 1



import deal


@deal.ensure(lambda _: _.result.startswith(_.left))
@deal.ensure(lambda _: _.result.endswith(_.right))
@deal.ensure(lambda _: len(_.result) == len(_.left) + len(_.right))
@deal.has()
def concat(left: str, right: str) -> str:
    """Concatenate 2 given strings.

    https://deal.readthedocs.io/basic/motivation.html
    """
    return left + right


test_concat = deal.cases(concat)



import codecs

from pythonfuzz.main import PythonFuzz

import deal


def encode(text: str) -> str:
    return codecs.encode(text, encoding='rot13')


@deal.ensure(lambda text, result: encode(result) == text)
def decode(text: str) -> str:
    assert text != 'bad'
    return codecs.encode(text, encoding='rot13')


def fuzz():
    test = deal.cases(decode)
    PythonFuzz(test)()


if __name__ == '__main__':
    fuzz()



import ast
from typing import Iterator, Optional

import astroid

from .._contract import Category
from .._stub import StubsManager
from .common import TOKENS, Extractor, Token, get_full_name, get_name, get_stub, infer
from .contracts import get_contracts
from .value import get_value


get_markers = Extractor()
DEFINITELY_RANDOM_FUNCS = frozenset({
    'randint',
    'randbytes',
    'randrange',
    'getrandbits',
    'shuffle',
})
SYSCALLS = frozenset({
    # https://docs.python.org/3/library/os.html#process-management
    'os.abort',
    'os.execv',
    'os.fork',
    'os.forkpty',
    'os.kill',
    'os.killpg',
    'os.plock',
    'os.posix_spawn',
    'os.posix_spawnp',
    'os.putenv',
    'os.startfile',
    'os.system',
    'os.wait',
    'os.wait3',
    'os.wait4',
    'os.waitid',
    'os.waitpid',

    'subprocess.call',
    'subprocess.check_call',
    'subprocess.check_out',
    'subprocess.getoutput',
    'subprocess.getstatusoutput',
    'subprocess.run',
    'subprocess.Popen',
})
SYSCALLS_PREFIXES = ('os.exec', 'os.spawn', 'os.popen')
TIMES = frozenset({
    'os.times',
    'datetime.now',
    'date.today',
    'datetime.datetime.now',
    'datetime.date.today',

    'time.clock_gettime',
    'time.clock_gettime_ns',
    'time.get_clock_info',
    'time.monotonic',
    'time.monotonic_ns',
    'time.perf_counter',
    'time.perf_counter_ns',
    'time.process_time',
    'time.process_time_ns',
    'time.time',
    'time.time_ns',
    'time.thread_time',
    'time.thread_time_ns',
})
NETWORK_FUNCS = frozenset({
    '_socket.socket.connect',
    '_socket.socket.sendall',
    '_socket.socket.recv',
    '_socket.socket.listen',
    '_socket.socket.bind',
    '_socket.socket.getaddrinfo',
    '_socket.socket.close',

    'asyncio.streams.open_connection',
    'asyncio.streams.start_server',
    'asyncio.streams.open_unix_connection',
    'asyncio.streams.start_unix_server',
})


@get_markers.register(*TOKENS.GLOBAL)
def handle_global(expr, **kwargs) -> Optional[Token]:
    return Token(marker='global', line=expr.lineno, col=expr.col_offset)


@get_markers.register(*TOKENS.NONLOCAL)
def handle_nonlocal(expr, **kwargs) -> Optional[Token]:
    return Token(marker='global', line=expr.lineno, col=expr.col_offset)


@get_markers.register(ast.Import)
def handle_ast_import(expr: ast.Import, **kwargs) -> Optional[Token]:
    return Token(marker='import', line=expr.lineno, col=expr.col_offset)


@get_markers.register(astroid.Import)
def handle_astroid_import(expr: astroid.Import, **kwargs) -> Optional[Token]:
    return Token(marker='import', line=expr.lineno, col=expr.col_offset)


@get_markers.register(ast.ImportFrom)
def handle_ast_import_from(expr: ast.ImportFrom, **kwargs) -> Optional[Token]:
    return Token(marker='import', line=expr.lineno, col=expr.col_offset)


@get_markers.register(astroid.ImportFrom)
def handle_astroid_import_from(expr: astroid.ImportFrom, **kwargs) -> Optional[Token]:
    return Token(marker='import', line=expr.lineno, col=expr.col_offset)


@get_markers.register(*TOKENS.CALL)
def handle_call(expr, dive: bool = True, stubs: Optional[StubsManager] = None) -> Iterator[Token]:
    name = get_name(expr.func)
    if name is None:
        return

    # stdout, stderr, stdin
    token = _check_print(expr=expr, name=name)
    if token is not None:
        yield token
        return
    if name.startswith('sys.stdout'):
        yield Token(marker='stdout', value='sys.stdout.')
        return
    if name.startswith('sys.stderr'):
        yield Token(marker='stderr', value='sys.stderr.')
        return
    if name.startswith('sys.stdin'):
        yield Token(marker='stdin', value='sys.stdin.')
        return
    if name == 'input':
        yield Token(marker='stdin', value='input')
        return

    # random, import,
    if name == '__import__':
        yield Token(marker='import')
        return
    if _is_random(expr=expr, name=name):
        yield Token(marker='random', value=name)
        return
    if _is_syscall(expr=expr, name=name):
        yield Token(marker='syscall', value=name)
        return
    if _is_time(expr=expr, name=name):
        yield Token(marker='time', value=name)
        return

    # read and write
    if name == 'open':
        if _is_open_to_write(expr):
            yield Token(marker='write', value='open')
        else:
            yield Token(marker='read', value='open')
        return
    if _is_pathlib_write(expr):
        yield Token(marker='write', value='Path.open')
        return

    yield from _infer_markers(expr=expr, dive=dive, stubs=stubs)


def _infer_markers(expr, dive: bool, stubs: Optional[StubsManager] = None) -> Iterator[Token]:
    inferred = infer(expr=expr.func)
    stubs_found = False
    if type(expr) is astroid.Call and stubs is not None:
        for token in _markers_from_stubs(expr=expr, inferred=inferred, stubs=stubs):
            stubs_found = True
            yield token

    if not stubs_found:
        for token in _markers_from_inferred(expr=expr, inferred=inferred):
            dive = False
            yield token
        if dive:
            yield from _markers_from_func(expr=expr, inferred=inferred)


def _markers_from_inferred(expr: astroid.NodeNG, inferred: tuple) -> Iterator[Token]:
    for node in inferred:
        module, full_name = get_full_name(node)
        qual_name = f'{module}.{full_name}'
        if qual_name in NETWORK_FUNCS:
            yield Token(
                marker='network',
                value=full_name,
                line=expr.lineno,
                col=expr.col_offset,
            )
            return
        if isinstance(node, astroid.BoundMethod):
            if node.bound.pytype() == 'random.Random':
                yield Token(
                    marker='random',
                    value=full_name,
                    line=expr.lineno,
                    col=expr.col_offset,
                )
                return


def _is_open_to_write(expr) -> bool:
    for arg in expr.args:
        if isinstance(arg, astroid.Const) and arg.value == 'w':
            return True
        if isinstance(arg, ast.Str) and 'w' in arg.s:
            return True

    if not expr.keywords:
        return False
    for arg in expr.keywords:
        if arg.arg != 'mode':
            continue
        if isinstance(arg.value, astroid.Const) and 'w' in arg.value.value:
            return True
        if isinstance(arg.value, ast.Str) and 'w' in arg.value.s:
            return True
    return False


def _is_pathlib_write(expr) -> bool:
    if not isinstance(expr, astroid.Call):
        return False
    if not isinstance(expr.func, astroid.Attribute):
        return False
    if expr.func.attrname not in ('write_text', 'write_bytes', 'open'):
        return False

    # if it's open, check that mode is "w"
    if expr.func.attrname == 'open':
        if not _is_open_to_write(expr):
            return False

    for value in infer(expr.func.expr):
        if isinstance(value, astroid.Instance):
            if value.pytype().startswith('pathlib.'):
                return True
    return False


def _markers_from_stubs(expr: astroid.Call, inferred, stubs: StubsManager) -> Iterator[Token]:
    for value in inferred:
        if type(value) is not astroid.FunctionDef:
            continue
        module_name, func_name = get_full_name(expr=value)
        stub = get_stub(module_name=module_name, expr=value, stubs=stubs)
        if stub is None:
            continue
        names = stub.get(func=func_name, contract=Category.HAS)
        for name in names:
            yield Token(marker=name, line=expr.lineno, col=expr.col_offset)


def _markers_from_func(expr: astroid.NodeNG, inferred: tuple) -> Iterator[Token]:
    for value in inferred:
        if not isinstance(value, (astroid.FunctionDef, astroid.UnboundMethod)):
            continue

        # recursively infer markers from the function body
        for token in get_markers(body=value.body, dive=False):
            yield Token(
                marker=token.marker,
                value=token.value,
                line=expr.lineno,
                col=expr.col_offset,
            )

        # get explicitly specified markers from `@deal.has`
        for contract in get_contracts(value):
            if contract.name != 'has':
                continue
            for arg in contract.args:
                value = get_value(arg)
                if type(value) is not str:
                    continue
                yield Token(marker=value, line=expr.lineno, col=expr.col_offset)
    return None


def _check_print(expr, name: str) -> Optional[Token]:
    """Return token if expr is `print` function call.

    Marker type depends on `file=` keyword argument.
    If it is missed, the type is `stdout`.
    If it is `stdout` or `stderr`, the type is `stdout` or `stderr`.
    Otherwise, there is no marker. It writes something into a stream, and it's ok.
    """
    if name != 'print':
        return None
    for kwarg in (expr.keywords or []):
        if kwarg.arg != 'file':
            continue
        value = get_name(expr=kwarg.value)
        if value in ('stdout', 'sys.stdout'):
            return Token(
                marker='stdout',
                value='print',
                line=expr.lineno,
                col=expr.col_offset,
            )
        if value in ('stderr', 'sys.stderr'):
            return Token(
                marker='stderr',
                value='print',
                line=expr.lineno,
                col=expr.col_offset,
            )
        return None
    return Token(
        marker='stdout',
        value='print',
        line=expr.lineno,
        col=expr.col_offset,
    )


def _is_random(expr, name: str) -> bool:
    if name.startswith('random.'):
        return True
    if '.' in name:
        return False
    if name in DEFINITELY_RANDOM_FUNCS:
        return True
    return False


def _is_syscall(expr, name: str) -> bool:
    if name in SYSCALLS:
        return True
    if name.startswith(SYSCALLS_PREFIXES):
        return True
    return False


def _is_time(expr, name: str) -> bool:
    if name in TIMES:
        return True
    if f'time.{name}' in TIMES:
        return True
    return False



from typing import TYPE_CHECKING, Dict, List

from . import introspection


if TYPE_CHECKING:
    from sphinx.application import Sphinx as SphinxApp
    from sphinx.ext.autodoc import Options


CLASSIC_CONTRACTS = (introspection.Pre, introspection.Post, introspection.Ensure)


def autodoc(app: 'SphinxApp') -> None:
    """
    Activate the hook for [sphinx] that includes contracts into
    documentation generated by [autodoc].

    [sphinx]: http://www.sphinx-doc.org/
    [autodoc]: https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html
    """
    assert 'sphinx.ext.autodoc' in app.extensions
    app.connect('autodoc-process-docstring', _process_docstring)


def _process_docstring(
    app: 'SphinxApp',
    what: str,
    name: str,
    obj,
    options: 'Options',
    lines: List[str],
) -> None:
    raises: Dict[str, str] = dict()
    contracts: List[str] = []
    examples: List[str] = []
    for contract in introspection.get_contracts(obj):
        if isinstance(contract, introspection.Raises):
            for exc in contract.exceptions:
                raises.setdefault(exc.__qualname__, '')
            continue

        if isinstance(contract, introspection.Reason):
            message = contract.message or f'``{contract.source}``'
            raises[contract.event.__qualname__] = message
            continue

        if isinstance(contract, CLASSIC_CONTRACTS):
            message = contract.message or f'``{contract.source}``'
            contracts.append(f'  * {message}')
            continue

        if isinstance(contract, introspection.Has):
            lines.append(':side-effects:')
            lines.extend(f'  * {m}' for m in contract.markers)
            continue

        if isinstance(contract, introspection.Example):
            examples.append(f'  * ``{contract.source}``')
            continue

        raise RuntimeError('unreachable')

    for exc_name, descr in sorted(raises.items()):
        lines.append(f':raises {exc_name}: {descr}')
    if contracts:
        lines.append(':contracts:')
        lines.extend(contracts)
    if examples:
        lines.append(':examples:')
        lines.extend(examples)



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['deal',
 'deal._cli',
 'deal._runtime',
 'deal.introspection',
 'deal.linter',
 'deal.linter._extractors']

package_data = \
{'': ['*'],
 'deal.introspection': ['.hypothesis/*',
                        '.hypothesis/unicode_data/*',
                        '.hypothesis/unicode_data/13.0.0/*',
                        '.mypy_cache/*',
                        '.mypy_cache/3.9/*',
                        '.mypy_cache/3.9/_pytest/*',
                        '.mypy_cache/3.9/_pytest/_code/*',
                        '.mypy_cache/3.9/_pytest/_io/*',
                        '.mypy_cache/3.9/_pytest/assertion/*',
                        '.mypy_cache/3.9/_pytest/config/*',
                        '.mypy_cache/3.9/_pytest/mark/*',
                        '.mypy_cache/3.9/_typeshed/*',
                        '.mypy_cache/3.9/asyncio/*',
                        '.mypy_cache/3.9/attr/*',
                        '.mypy_cache/3.9/collections/*',
                        '.mypy_cache/3.9/concurrent/*',
                        '.mypy_cache/3.9/concurrent/futures/*',
                        '.mypy_cache/3.9/ctypes/*',
                        '.mypy_cache/3.9/deal/*',
                        '.mypy_cache/3.9/deal/_runtime/*',
                        '.mypy_cache/3.9/deal/introspection/*',
                        '.mypy_cache/3.9/distutils/*',
                        '.mypy_cache/3.9/docutils/*',
                        '.mypy_cache/3.9/docutils/languages/*',
                        '.mypy_cache/3.9/docutils/parsers/*',
                        '.mypy_cache/3.9/docutils/parsers/rst/*',
                        '.mypy_cache/3.9/docutils/readers/*',
                        '.mypy_cache/3.9/docutils/transforms/*',
                        '.mypy_cache/3.9/docutils/utils/*',
                        '.mypy_cache/3.9/docutils/writers/*',
                        '.mypy_cache/3.9/email/*',
                        '.mypy_cache/3.9/html/*',
                        '.mypy_cache/3.9/http/*',
                        '.mypy_cache/3.9/hypothesis/*',
                        '.mypy_cache/3.9/hypothesis/extra/*',
                        '.mypy_cache/3.9/hypothesis/extra/django/*',
                        '.mypy_cache/3.9/hypothesis/internal/*',
                        '.mypy_cache/3.9/hypothesis/internal/conjecture/*',
                        '.mypy_cache/3.9/hypothesis/internal/conjecture/dfa/*',
                        '.mypy_cache/3.9/hypothesis/internal/conjecture/shrinking/*',
                        '.mypy_cache/3.9/hypothesis/strategies/*',
                        '.mypy_cache/3.9/hypothesis/strategies/_internal/*',
                        '.mypy_cache/3.9/hypothesis/utils/*',
                        '.mypy_cache/3.9/hypothesis/vendor/*',
                        '.mypy_cache/3.9/importlib/*',
                        '.mypy_cache/3.9/importlib_metadata/*',
                        '.mypy_cache/3.9/importlib_resources/*',
                        '.mypy_cache/3.9/iniconfig/*',
                        '.mypy_cache/3.9/jinja2/*',
                        '.mypy_cache/3.9/json/*',
                        '.mypy_cache/3.9/logging/*',
                        '.mypy_cache/3.9/markupsafe/*',
                        '.mypy_cache/3.9/multiprocessing/*',
                        '.mypy_cache/3.9/numpy/*',
                        '.mypy_cache/3.9/numpy/compat/*',
                        '.mypy_cache/3.9/numpy/core/*',
                        '.mypy_cache/3.9/numpy/distutils/*',
                        '.mypy_cache/3.9/numpy/fft/*',
                        '.mypy_cache/3.9/numpy/lib/*',
                        '.mypy_cache/3.9/numpy/linalg/*',
                        '.mypy_cache/3.9/numpy/ma/*',
                        '.mypy_cache/3.9/numpy/matrixlib/*',
                        '.mypy_cache/3.9/numpy/polynomial/*',
                        '.mypy_cache/3.9/numpy/random/*',
                        '.mypy_cache/3.9/numpy/testing/*',
                        '.mypy_cache/3.9/numpy/typing/*',
                        '.mypy_cache/3.9/os/*',
                        '.mypy_cache/3.9/packaging/*',
                        '.mypy_cache/3.9/py/*',
                        '.mypy_cache/3.9/pytest/*',
                        '.mypy_cache/3.9/sphinx/*',
                        '.mypy_cache/3.9/sphinx/builders/*',
                        '.mypy_cache/3.9/sphinx/directives/*',
                        '.mypy_cache/3.9/sphinx/domains/*',
                        '.mypy_cache/3.9/sphinx/environment/*',
                        '.mypy_cache/3.9/sphinx/environment/adapters/*',
                        '.mypy_cache/3.9/sphinx/environment/collectors/*',
                        '.mypy_cache/3.9/sphinx/ext/*',
                        '.mypy_cache/3.9/sphinx/ext/autodoc/*',
                        '.mypy_cache/3.9/sphinx/locale/*',
                        '.mypy_cache/3.9/sphinx/pycode/*',
                        '.mypy_cache/3.9/sphinx/search/*',
                        '.mypy_cache/3.9/sphinx/testing/*',
                        '.mypy_cache/3.9/sphinx/transforms/*',
                        '.mypy_cache/3.9/sphinx/util/*',
                        '.mypy_cache/3.9/sphinx/util/stemmer/*',
                        '.mypy_cache/3.9/toml/*',
                        '.mypy_cache/3.9/typeguard/*',
                        '.mypy_cache/3.9/unittest/*',
                        '.mypy_cache/3.9/urllib/*',
                        '.mypy_cache/3.9/zoneinfo/*'],
 'deal.linter': ['stubs/*',
                 'stubs/aiohttp/*',
                 'stubs/cpython/*',
                 'stubs/cpython/asyncio/*',
                 'stubs/cpython/http/*',
                 'stubs/cpython/sqlite3/*',
                 'stubs/cpython/urllib/*',
                 'stubs/dateutil/*',
                 'stubs/dateutil/parser/*',
                 'stubs/dateutil/tz/*',
                 'stubs/dateutil/zoneinfo/*',
                 'stubs/marshmallow/*',
                 'stubs/pytz/*',
                 'stubs/requests/*',
                 'stubs/urllib3/*',
                 'stubs/urllib3/contrib/*',
                 'stubs/urllib3/util/*']}

install_requires = \
['astroid', 'deal-solver', 'hypothesis', 'pygments', 'typeguard', 'vaa>=0.2.1']

extras_require = \
{'docs': ['m2r2', 'myst-parser', 'sphinx==3.5.*', 'sphinx-rtd-theme==0.5.*'],
 'lint': ['flake8',
          'flake8-commas',
          'flake8-quotes',
          'mypy>=0.900',
          'mypy_test>=0.1.1',
          'isort'],
 'test': ['coverage[toml]',
          'pytest-cov',
          'isort',
          'marshmallow',
          'pytest',
          'urllib3',
          'sphinx']}

entry_points = \
{'flake8.extension': ['DEAL = deal.linter:Checker']}

setup(name='deal',
      version='4.19.2',
      description='**Deal** is a Python library for [design by contract][wiki] (DbC) programming.',
      author='Gram',
      author_email='gram@orsinium.dev',
      url='https://github.com/life4/deal',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      entry_points=entry_points,
      python_requires='>=3.7',
     )



import re
from pathlib import Path
from textwrap import dedent

import pytest

import deal
from deal._cli._main import get_commands
from deal.linter._rules import CheckMarkers, rules


root = Path(__file__).parent.parent / 'docs'
rex_code = re.compile(r'```python\s+run\n(.+?)\n\s*```', re.DOTALL)


def get_code_blocks(path: Path):
    content = path.read_text()
    for match in rex_code.finditer(content):
        yield dedent(match.group(1))


def test_all_codes_listed():
    path = root / 'basic' / 'linter.md'
    content = path.read_text()
    for checker in rules:
        if type(checker) is CheckMarkers:
            continue
        line = '| DEAL{:03d} |'.format(checker.code)
        assert line in content

    for marker, code in CheckMarkers.codes.items():
        line = '| DEAL{:03d} | missed marker ({})'.format(code, marker)
        assert line in content


def test_all_marker_codes_listed():
    path = root / 'basic' / 'side-effects.md'
    content = path.read_text()
    lines = content.splitlines()
    for marker, code in CheckMarkers.codes.items():
        code_cell = '| DEAL{:03d} | '.format(code)
        assert code_cell in content
        for line in lines:
            if code_cell in line:
                assert marker in line
                break
        else:
            assert False, 'missing marker {}'.format(marker)


def test_cli_included():
    path = root / 'details' / 'cli.md'
    content = path.read_text()
    for name, cmd in get_commands().items():
        # has header
        tmpl = '## {n}\n\n'
        line = tmpl.format(n=name)
        assert line in content

        # has autodoc
        tmpl = '```{{eval-rst}}\n.. autofunction:: deal._cli._{n}.{c}\n```'
        line = tmpl.format(n=name, c=cmd.__name__)
        assert line in content

        # header and autodoc go next to each other
        tmpl = '## {n}\n\n```{{eval-rst}}\n.. autofunction:: deal._cli._{n}.{c}\n```'
        line = tmpl.format(n=name, c=cmd.__name__)
        assert line in content


def test_examples_included():
    path = root / 'details' / 'examples.md'
    content = path.read_text()
    for path in root.parent.joinpath('examples').iterdir():
        if '__' in path.name:
            continue
        # has header
        tmpl = '## {}\n\n'
        line = tmpl.format(path.stem)
        assert line in content

        # has include
        tmpl = '```{{eval-rst}}\n.. literalinclude:: ../../examples/{}\n```'
        line = tmpl.format(path.name)
        assert line in content


@pytest.mark.parametrize('name', deal.__all__)
def test_all_public_have_docstring(name: str):
    obj = getattr(deal, name)
    doc = obj.__doc__
    assert doc is not None, f'{name} has no docstring'


def test_all_public_listed_in_refs():
    path = root / 'basic' / 'refs.md'
    content = path.read_text()
    for name in deal.__all__:
        if name[0] == '_':
            continue
        if name in {'introspection', 'Scheme', 'TestCase'}:
            continue
        assert f'deal.{name}`' in content


def test_all_cli_commands_listed_in_refs():
    path = root / 'basic' / 'refs.md'
    content = path.read_text()
    for name, cmd in get_commands().items():
        assert f'`{name}`' in content
        assert f'`details/cli:{name}`' in content


def test_all_public_listed_in_api():
    path = root / 'details' / 'api.md'
    content = path.read_text()
    for name in deal.__all__:
        if name[0] == '_':
            continue
        if name in {'Scheme'}:
            continue
        assert f':: deal.{name}' in content


@pytest.mark.parametrize(
    'path',
    [pytest.param(p, id=p.name) for p in root.glob('**/*.md')],
)
def test_code_snippets(path: Path):
    for code in get_code_blocks(path):
        exec(code, dict(
            deal=deal,
            pytest=pytest,
        ))



from deal.linter._error import Error


def test_error():
    err = Error(row=1, col=2, code=3, text='check')
    assert tuple(err) == (1, 2, 'DEAL003 check')
    assert str(err) == 'DEAL003 check'
    assert repr(err).startswith('Error(')



import deal


@deal.raises(ZeroDivisionError)
@deal.reason(ZeroDivisionError, lambda _: _.right == 0)
@deal.has()
def div1(left: float, right: float) -> float:
    """
    This implementation allows zero to be passed
    but raises ZeroDivisionError in that case.
    """
    return left / right


@deal.pre(lambda _: _.right != 0)
@deal.has()
def div2(left: float, right: float) -> float:
    """
    This implementation doesn't allow zero to be passed in a function.
    If it is accidentally passed, PreConditionError will be raised
    and the funcrion won't be executed.
    """
    return left / right


test_div1 = deal.cases(div1)
test_div2 = deal.cases(div2)



from typing import Optional

from .common import TOKENS, Extractor, Token, traverse
from .value import UNKNOWN, get_value


get_returns = Extractor()


def has_returns(body: list) -> bool:
    for expr in traverse(body=body):
        if isinstance(expr, TOKENS.RETURN + TOKENS.YIELD):
            return True
    return False


@get_returns.register(*TOKENS.RETURN)
def handle_return(expr) -> Optional[Token]:
    value = get_value(expr=expr.value)
    if value is UNKNOWN:
        return None
    return Token(value=value, line=expr.lineno, col=expr.value.col_offset)


@get_returns.register(*TOKENS.YIELD)
def handle_yield(expr) -> Optional[Token]:
    value = get_value(expr=expr.value)
    if value is UNKNOWN:
        return None
    return Token(value=value, line=expr.lineno, col=expr.value.col_offset)



import pytest

import deal

from .helpers import run_sync


@pytest.mark.parametrize('correct,incorrect', [(1, -1), (2, -2), (3, -3), (5, -5), (7, -7), (11, -11)])
def test_pre_contract_fulfilled(correct, incorrect):
    func = deal.pre(lambda x: x > 0)(lambda x: x)
    assert func(correct) == correct
    with pytest.raises(deal.PreContractError):
        func(incorrect)


@pytest.mark.parametrize('correct,incorrect_min,incorrect_max',
                         [(1, -1, 20), (2, -2, 21), (3, -3, 22), (5, -5, 23), (7, -7, 24), (9, -11, 25)])
def test_chain_all_contracts_fulfilled(correct, incorrect_min, incorrect_max):
    func = deal.pre(lambda x: x < 10)(lambda x: x)
    func = deal.pre(lambda x: x > 0)(func)
    assert func(correct) == correct
    with pytest.raises(deal.PreContractError):
        func(incorrect_min)
    with pytest.raises(deal.PreContractError):
        func(incorrect_max)


def test_correct_exceptions_raised_on_contract_fail():
    func = deal.pre(lambda x: x > 0)(lambda x: x)
    with pytest.raises(deal.PreContractError):
        func(-2)

    func = deal.pre(lambda x: x > 0, message='TEST')(lambda x: x)
    try:
        func(-2)
    except AssertionError as e:
        assert e.args[0] == 'TEST'

    func = deal.pre(lambda x: x > 0, exception=NameError)(lambda x: x)
    with pytest.raises(NameError):
        func(-2)

    func = deal.pre(lambda x: x > 0, exception=NameError('TEST'))(lambda x: x)
    with pytest.raises(NameError):
        func(-2)
    try:
        func(-2)
    except NameError as e:
        assert e.args[0] == 'TEST'

    func = deal.pre(lambda x: x > 0, message='TEST', exception=NameError)(lambda x: x)
    with pytest.raises(NameError):
        func(-2)
    try:
        func(-2)
    except NameError as e:
        assert e.args[0] == 'TEST'


def test_raise_error_with_param_on_contract_failure():
    func = deal.pre(lambda x: x > 0 or 'TEST')(lambda x: x)
    assert func(4) == 4

    with pytest.raises(deal.PreContractError):
        func(-2)

    try:
        func(-2)
    except deal.PreContractError as e:
        assert e.args[0] == 'TEST'


def test_method_decoration_name_is_correct():
    @deal.pre(lambda x: x > 0)
    def some_function(x):
        return x

    assert some_function.__name__ == 'some_function'


def test_class_method_decorator_raises_error_on_contract_fail():
    class Class:
        y = 7

        @deal.pre(lambda self, x: x > 0)
        def method(self, x):
            return x * 2

        @deal.pre(lambda self, x: x > 0)
        def method2(self, y):
            return self.y

    assert Class().method(2) == 4
    assert Class().method2(2) == 7
    with pytest.raises(deal.PreContractError):
        Class().method(-2)
    with pytest.raises(deal.PreContractError):
        Class().method2(-2)


def test_contract_returns_message():
    func = deal.pre(lambda x: x > 0 or 'TEST')(lambda x: x)
    assert func(4) == 4

    with pytest.raises(deal.PreContractError):
        func(-2)

    try:
        func(-2)
    except deal.PreContractError as e:
        assert e.args[0] == 'TEST'


def test_text_from_contract_rewrites_default_one():
    @deal.pre(lambda x: x > 0 or 'new message', message='old message')
    def double(x):
        return x * 2

    try:
        double(-1)
    except deal.PreContractError as exc:
        assert exc.args[0] == 'new message'


def test_decorating_async_function():
    @deal.pre(lambda x: x > 0)
    async def double(x):
        return x * 2

    assert run_sync(double(2)) == 4
    with pytest.raises(deal.PreContractError):
        run_sync(double(-2))


def test_decorating_generator():
    @deal.pre(lambda x: x > 0)
    def double(x):
        yield x
        yield x * 2
        yield x * 4

    assert list(double(2)) == [2, 4, 8]
    with pytest.raises(deal.PreContractError):
        list(double(-2))
